# -*- coding: utf-8 -*-
"""
High level functions to make complex tasks easier.

Functions
---------
parapply
    Run a function on a pandas DataFrame in parralel on a cluster
parapply_summary
    Run parapply and merge the results (don't concatenate).
splitrun
    Split a file, run a command in parallel, return result.
"""
import os as _os
from six import text_type as _txt
from six import string_types as _str
from six import integer_types as _int

###############################################################################
#                               Import Ourself                                #
###############################################################################

from . import run as _run
from . import conf as _conf
from . import logme as _logme
from . import batch_systems as _batch
from .job import Job as _Job

_options = _batch.options

###############################################################################
#                       Try Import Non-Required Modules                       #
###############################################################################

try:
    import numpy as _np
    import pandas as _pd
except ImportError:
    _logme.log('Could not import numpy and pandas for helpers', 'debug')

__all__ = ['jobify', 'parapply', 'parapply_summary', 'splitrun']

###############################################################################
#                                 Decorators                                  #
###############################################################################


def jobify(name=None, profile=None, qtype=None, submit=True, **kwds):
    """Decorator to make any function a job.

    Will make any function return a Job object that will execute the function
    on the cluster.

    If `submit` is `True`, the job will be submitted when it is returned.

    Usage:

    .. code:: python

        @fyrd.jobify(name='my_job', profile='small', mem='8GB',
                     time='00:10:00', imports=['from time import sleep'])
        def do_something(file_path, iteration_count=24):
            for i in range(iteration_count):
                print(file_path + i)
                sleep(1)
            return file_path

        job = do_something('my_file.txt')
        out = job.get()

    Parameters
    ----------
    name : str, optional
        Optional name of the job. If not defined, guessed. If a job of the same
        name is already queued, an integer job number (not the queue number)
        will be added, ie.  <name>.1
    qtype : str, optional
        Override the default queue type
    profile : str, optional
        The name of a profile saved in the conf
    submit : bool, optional
        Submit the Job before returning it
    kwds
        *All other keywords are parsed into cluster keywords by the options
        system.* For available keywords see `fyrd.option_help()`

    Returns
    -------
    fyrd.job.Job
        A Job class initialized with the decorated function.

    Examples
    --------

    >>> import fyrd
    >>> @fyrd.jobify(name='test_job', mem='1GB')
    ... def test(string, iterations=4):
    ...     \"\"\"This does basically nothing!\"\"\"
    ...     outstring = ""
    ...     for i in range(iterations):
    ...         outstring += "Version {0}: {1}".format(i, string)
    ...     return outstring
    >>> j = test('hi')
    >>> j.get()
    'Version 0: hiVersion 1: hiVersion 2: hiVersion 3: hiVersion 4: hi'
    """
    def deco(func):
        """This will be the actual decorator for the function."""
        # Munge the job docstring to preserve function documentation
        fdoc = func.__doc__
        if fdoc:
            fdoc = "\nOriginal Docstring:\n\n{0}".format(fdoc)
        else:
            fdoc = ""
        def wrapper(*args, **kwargs):
            """This will convert the function into a Job object."""
            job = _Job(func, args=args, kwargs=kwargs, name=name, qtype=qtype,
                       profile=profile, **kwds)
            jdoc = job.__doc__
            if jdoc:
                jdoc = "\n\nJob documentation:\n\n{0}".format(jdoc)
            else:
                jdoc = ""
            job.__doc__ = (
                "This is a fyrd.job.Job decorated function, " +
                "to get results run `.get()`.\n" + fdoc + jdoc
            )
            if submit:
                job.submit()
            return job
        wrapper.__doc__ = (
            "This is a fyrd.job.Job decorated function.\n" +
            "\nWhen you call it it will return a Job object from " +
            "which you can get\nthe results with the `.get()` method.\n" +
            fdoc
        )
        return wrapper
    return deco


###############################################################################
#                                  parapply                                   #
###############################################################################


def parapply(jobs, df, func, args=(), profile=None, applymap=False,
             merge_axis=0, merge_apply=False, name='parapply', imports=None,
             direct=True, **kwds):
    """Split a dataframe, run apply in parallel, return result.

    This function will split a dataframe into however many pieces are requested
    with the jobs argument, run apply in parallel by submitting the jobs to the
    cluster, and then recombine the outputs.

    If the 'clean_files' and 'clean_outputs' arguments are not passed, we
    delete all intermediate files and output files by default.

    This function will take any keyword arguments accepted by Job, which can
    be found by running fyrd.options.option_help(). It also accepts any of
    the keywords accepted by by pandas.DataFrame.apply(), found
    `here <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.apply.html>`_

    Parameters
    ----------
    jobs : int
        Number of pieces to split the dataframe into
    df : DataFrame
        Any pandas DataFrame
    args : tuple
        Positional arguments to pass to the function, keyword arguments can
        just be passed directly.
    profile : str
        A fyrd cluster profile to use
    applymap : bool
        Run applymap() instead of apply()
    merge_axis : int
        Which axis to merge on, 0 or 1, default is 1 as apply transposes
        columns
    merge_apply : bool
        Apply the function on the merged dataframe also
    name : str
        A prefix name for all of the jobs
    imports : list
        A list of imports in any format, e.g.
        `['import numpy', 'scipy', 'from numpy import mean']`
    direct : bool
        Whether to run the function directly or to return a Job. Default True.

    Any keyword arguments recognized by fyrd will be used for job
    submission.

    *Additional keyword arguments will be passed to DataFrame.apply()*

    Returns
    -------
    DataFrame
        A recombined DataFrame: concatenated version of original split
        DataFrame

    Example
    -------
    >>> import numpy
    >>> import pandas
    >>> import fyrd
    >>> df = pandas.DataFrame([[0, 1], [2, 6], [9, 24], [13, 76], [4, 12]])
    >>> df['sum'] = fyrd.helpers.parapply(2, df, lambda x: x[0]+x[1], axis=1)
    >>> df
        0   1  sum
        0   0   1    1
        1   2   6    8
        2   9  24   33
        3  13  76   89
        4   4  12   16

    See Also
    --------
    parapply_summary: Merge results of parapply using applied function
    splitrun: Run a command in parallel on a split file
    """
    if direct:
        return _parapply(
            jobs, df, func, args=args, profile=profile, applymap=applymap,
            merge_axis=merge_axis, merge_apply=merge_apply, name=name,
            imports=imports, **kwds
        )
    kwargs = dict(
        args=args, profile=profile, applymap=applymap,
        merge_axis=merge_axis, merge_apply=merge_apply, name=name,
        imports=imports
    )
    kwds = _options.sanitize_arguments(kwds)
    kwargs.update(kwds)
    kwargs['imports']  = _run.get_all_imports(func, kwargs)
    kwargs['syspaths'] = _run.update_syspaths(func, kwargs)
    return _wrap_runner(
        _parapply,
        *(jobs, df, func),
        **kwargs
    )


def _parapply(jobs, df, func, args=(), profile=None, applymap=False,
              merge_axis=0, merge_apply=False, name='parapply', imports=None,
              **kwds):
    """Direct running function for parapply, see parapply docstring."""
    # Handle arguments
    if not isinstance(jobs, _int):
        raise ValueError('Jobs argument must be an integer.')
    if not isinstance(df, (_pd.core.frame.DataFrame, _np.ndarray)):
        raise ValueError('df must be a dataframe or numpy array, is {}'
                         .format(type(df)))
    if not callable(func):
        raise ValueError('function must be callable, current type is {}'
                         .format(type(func)))
    if profile is not None and not isinstance(profile, (_str, _txt)):
        raise ValueError('Profile must be a string, is {}'
                         .format(type(profile)))
    fyrd_kwds, pandas_kwds = _options.split_keywords(kwds)

    # Get name
    name = name if name else 'split_file'

    # Set up auto-cleaning
    if 'clean_files' not in fyrd_kwds:
        fyrd_kwds['clean_files'] = True
    if 'clean_outputs' not in fyrd_kwds:
        fyrd_kwds['clean_outputs'] = True

    # Split dataframe
    _logme.log('Splitting dataframe', 'debug')
    dfs = _np.array_split(df, jobs)
    assert len(dfs) == jobs

    # Pick function
    sub_func = _run_applymap if applymap else _run_apply

    # Some sane imports
    imports  = _run.listify(imports) if imports else []
    imports += ['import numpy as np', 'import numpy', 'import scipy as sp',
                'import sp', 'import pandas as pd', 'import pandas',
                'from matplotlib import pyplot as plt',
                'from scipy import stats']
    imports = _run.export_imports(func, {'imports': imports})

    # Run the functions
    _logme.log('Submitting jobs', 'debug')
    outs = []
    count = 1
    for d in dfs:
        nm = '{}_{}_of_{}'.format(name, count, jobs)
        outs.append(
            _Job(sub_func, (d, func, args, pandas_kwds),
                 profile=profile, name=nm, imports=imports,
                 **fyrd_kwds).submit()
        )
        count += 1

    # Get the results
    _logme.log('Waiting for results', 'debug')
    results = []
    for out in outs:
        try:
            results.append(out.get())
        except IOError:
            _logme.log('Result getting failed, most likely one of the child ' +
                       'jobs crashed, check the error files', 'critical')
            raise

    # Return the recombined DataFrame
    _logme.log('Done, joinging', 'debug')
    try:
        out = _pd.concat(results, axis=merge_axis)
    except ValueError:
        _logme.log('DataFrame concatenation failed!', 'critical')
        raise

    if merge_apply:
        out = sub_func(out, func, args, pandas_kwds)

    return out


def parapply_summary(jobs, df, func, args=(), profile=None, applymap=False,
                     name='parapply', imports=None, direct=True, **kwds):
    """Run parapply for a function with summary stats.

    Instead of returning the concatenated result, merge the result using
    the same function as was used during apply.

    This works best for summary functions like `.mean()`, which do a linear
    operation on a whole dataframe or series.

    Parameters
    ----------
    jobs : int
        Number of pieces to split the dataframe into
    df : DataFrame
        Any pandas DataFrame
    args : tuple
        Positional arguments to pass to the function, keyword arguments can
        just be passed directly.
    profile : str
        A fyrd cluster profile to use
    applymap : bool
        Run applymap() instead of apply()
    merge_axis : int
        Which axis to merge on, 0 or 1, default is 1 as apply transposes
        columns
    merge_apply : bool
        Apply the function on the merged dataframe also
    name : str
        A prefix name for all of the jobs
    imports : list
        A list of imports in any format, e.g.
        `['import numpy', 'scipy', 'from numpy import mean']`
    direct : bool
        Whether to run the function directly or to return a Job. Default True.

    Any keyword arguments recognized by fyrd will be used for job
    submission.

    *Additional keyword arguments will be passed to DataFrame.apply()*

    Returns
    -------
    DataFrame
        A recombined DataFrame

    Example
    -------
    >>> import numpy
    >>> import pandas
    >>> import fyrd
    >>> df = pandas.DataFrame([[0, 1], [2, 6], [9, 24], [13, 76], [4, 12]])
    >>> df = fyrd.helpers.parapply_summary(2, df, numpy.mean)
    >>> df
    0     6.083333
    1    27.166667
    dtype: float64

    See Also
    --------
    parapply: Run a command in parallel on a DataFrame without merging the
    result
    """
    if direct:
        return _parapply_summary(
            jobs, df, func, args=args, profile=profile, applymap=applymap,
            name=name, imports=imports, **kwds
        )
    kwargs = dict(
        args=args, profile=profile, applymap=applymap,
        name=name, imports=imports
        )
    kwds = _options.sanitize_arguments(kwds)
    kwargs.update(kwds)
    kwargs['imports']  = _run.get_all_imports(func, kwargs)
    kwargs['syspaths'] = _run.update_syspaths(func, kwargs)
    return _wrap_runner(
        _parapply_summary,
        *(jobs, df, func),
        **kwargs
    )


def _parapply_summary(jobs, df, func, args=(), profile=None, applymap=False,
                      name='parapply', imports=None, **kwds):
    """Direct running function for parapply_sumary, see that docstring."""
    imports = _run.export_imports(func, {'imports': imports})
    out = parapply(jobs, df, func, args, profile, applymap, name=name,
                   merge_axis=1, imports=imports, **kwds)

    # Pick function and get args
    sub_func = _run_applymap if applymap else _run_apply
    pandas_kwds = _options.split_keywords(kwds)[1]

    # Transpose
    out = out.T

    # Run the function again
    return sub_func(out, func, args, pandas_kwds)


def _run_apply(df, func, args=None, pandas_kwds=None):
    """Run DataFrame.apply().

    Parameters
    ----------
    df : DataFrame
        Any pandas DataFrame
    args : tuple
        A tuple of arguments to submit to the function
    pandas_kwds : dict
        A dictionary of keyword arguments to pass to DataFrame.apply()

    Returns
    -------
    DataFrame
        The result of apply()
    """
    apply_kwds = {'args': args} if args else {}
    if pandas_kwds:
        apply_kwds.update(pandas_kwds)
    return df.apply(func, **apply_kwds)


def _run_applymap(df, func, args=None, pandas_kwds=None):
    """Run DataFrame.applymap().

    Parameters
    ----------
    df : DataFrame
        Any pandas DataFrame
    args : tuple
        A tuple of arguments to submit to the function
    pandas_kwds : dict
        A dictionary of keyword arguments to pass to DataFrame.apply()

    Returns
    -------
    DataFrame
        The result of apply()
    """
    apply_kwds = {'args': args} if args else {}
    if pandas_kwds:
        apply_kwds.update(pandas_kwds)
    return df.applymap(func, **apply_kwds)


###############################################################################
#                                  Split Run                                  #
###############################################################################


def splitrun(jobs, infile, inheader, command, args=None, kwargs=None,
             name=None, qtype=None, profile=None, outfile=None,
             outheader=False, merge_func=None, direct=True, **kwds):
    """Split a file, run command in parallel, return result.

    This function will split a file into however many pieces are requested
    with the jobs argument, and run command on each.

    Accepts exactly the same arguments as the Job class, with the exception of
    the first three and last four arguments, which are:

    - the number of jobs
    - the file to work on
    - whether the input file has a header
    - an optional output file
    - whether the output file has a header
    - an optional function to use to merge the resulting list, only used
      if there is no outfile.
    - whether to run directly or to return a Job. If direct is True, this
      function will just run and thus block until complete, if direct is
      False, the function will submit as a Job and return that Job.

    **Note**: If command is a string, `.format(file={file})` will be called on
    it, where file is each split file. If command is a function, the there must
    be an argument in either args or kwargs that contains `{file}`. It will be
    replaced with the *path to the file*, again by the format command.

    If outfile is specified, there must also be an '{outfile}' line in any
    script or an '{outfile}' argument in either args or kwargs. When this
    function completes, the file at outfile will contain the concatenated
    output files of all of the jobs.

    If the 'clean_files' and 'clean_outputs' arguments are not passed, we
    delete all intermediate files and output files by default.

    The intermediate files will be stored in the 'scriptpath' directory.

    Any header line is kept at the top of the file.

    Primary return value varies and is decided in this order:

    If outfile:
        the absolute path to that file

    If merge_func:
        the result of merge_func(list), where list is the list of outputs.

    Else:
        a list of results

    If direct is False, this function returns a fyrd.job.Job object which
    will return the results described above on get().

    Parameters
    ----------
    jobs : int
        Number of pieces to split the dataframe into
    infile : str
        The path to the file to be split.
    inheader : bool
        Does the input file have a header?
    command : function/str
        The command or function to execute.
    args : tuple/dict
        Optional arguments to add to command, particularly useful for
        functions.
    kwargs : dict
        Optional keyword arguments to pass to the command, only used for
        functions.
    name : str
        Optional name of the job. If not defined, guessed. If a job of the same
        name is already queued, an integer job number (not the queue number)
        will be added, ie.  <name>.1
    qtype : str
        Override the default queue type
    profile : str
        The name of a profile saved in the conf
    outfile : str
        The path to the expected output file.
    outheader : bool
        Does the input outfile have a header?
    merge_func : function
        An optional function used to merge the output list if there is no
        outfile.
    direct : bool
        Whether to run the function directly or to return a Job. Default True.

    *All other keywords are parsed into cluster keywords by the options
    system. For available keywords see `fyrd.option_help()` *

    Returns
    -------
    Varies
        See description above
    """
    kwds = _options.check_arguments(kwds)
    if direct:
        return _splitrun(
            jobs, infile, inheader, command, args=args, kwargs=kwargs,
            name=name, qtype=qtype, profile=profile, outfile=outfile,
            outheader=outheader, merge_func=merge_func, **kwds
        )
    else:
        kk = dict(
            args=args, kwargs=kwargs, name=name, qtype=qtype, profile=profile,
            outfile=outfile, outheader=outheader, merge_func=merge_func
        )
        kwds = _options.sanitize_arguments(kwds)
        kk.update(_options.check_arguments(kwds))
        if callable(command):
            kk['imports']  = _run.get_all_imports(command, kk)
            kk['syspaths'] = _run.update_syspaths(command, kk)
        return _wrap_runner(
            _splitrun,
            *(jobs, infile, inheader, command),
            **kk
        )


def _splitrun(jobs, infile, inheader, command, args=None, kwargs=None,
              name=None, qtype=None, profile=None, outfile=None,
              outheader=False, merge_func=None, **kwds):
    """This is the direct running function for `splitrun()`.

    Please see that function's docstring for information and do not call this
    function directly.
    """
    # Handle arguments
    if not isinstance(jobs, _int):
        raise ValueError('Jobs argument must be an integer.')
    if profile is not None and not isinstance(profile, (_str, _txt)):
        raise ValueError('Profile must be a string, is {}'
                         .format(type(profile)))
    kwds = _options.check_arguments(kwds)

    # Get name
    name = name if name else 'split_file'

    # Check file
    infile = _os.path.abspath(_os.path.expandvars(_os.path.expanduser(infile)))
    if not _os.path.isfile(infile):
        raise OSError('Cannot find file {}'.format(infile))

    # Set up auto-cleaning
    if 'clean_files' not in kwds:
        kwds['clean_files'] = True
    if 'clean_outputs' not in kwds:
        kwds['clean_outputs'] = True

    # Get script path, returns: kwds, runpath, outpath, scriptpath
    cpath = _conf.get_job_paths(kwds)[3]

    # Prep function
    if callable(command):
        # Add all imports from the function file to globals
        kwds['imports'] = _run.export_imports(command, kwds)

    # Split file
    _logme.log('Splitting file', 'debug')
    files = _run.split_file(infile, jobs, outpath=cpath, keep_header=inheader)
    assert len(files) == jobs

    # Run the functions
    _logme.log('Submitting jobs', 'debug')
    outs = []
    count = 1
    for f in files:
        if outfile:
            o = f + '.out'
        nm = '{}_{}_of_{}'.format(name, count, jobs)
        if callable(command):
            try:
                runargs, runkwargs = _run.replace_argument(
                    [args, kwargs], '{file}', f)
            except ValueError:
                raise ValueError("No '{file}' argument in either args or " +
                                 "kwargs")
            if outfile:
                runargs, runkwargs = _run.replace_argument(
                    [runargs, runkwargs], '{outfile}', o)
            outs.append(
                _Job(command, args=runargs, kwargs=runkwargs, name=nm,
                     qtype=qtype, profile=profile, **kwds).submit()
            )
        else:
            if outfile:
                cmnd = command.format(file=f, outfile=o)
            else:
                cmnd = command.format(file=f)
            outs.append(
                _Job(cmnd, name=nm, qtype=qtype, profile=profile,
                     **kwds).submit()
            )
        count += 1

    # Get the results
    _logme.log('Waiting for results', 'debug')
    results = []
    for out in outs:
        try:
            results.append(out.get())
        except IOError:
            _logme.log('Result getting failed, most likely one of the child ' +
                       'jobs crashed, check the error files', 'critical')
            raise

    # Delete intermediate files
    _logme.log('Removing intermediate files', 'debug')
    for f in files:
        assert _os.path.isfile(f)
        _os.remove(f)

    # Return the recombined DataFrame
    _logme.log('Done, joining', 'debug')
    if outfile:
        with open(outfile, 'w') as fout:
            if outheader:
                with open(files[0]) as fin:
                    fout.write(fin.readline())
            for f in files:
                with open(f + '.out') as fin:
                    if outheader:
                        fin.readline()
                    fout.write(fin.read())
        out = _os.path.abspath(outfile)

    elif merge_func:
        assert callable(merge_func)
        out = merge_func(results)

    elif isinstance(results[0], list):
        out = _run.merge_lists(results)

    else:
        out = results

    return out


###############################################################################
#                              Helper Functions                               #
###############################################################################


def _wrap_runner(command, *args, **kkk):
    """Run a command as a job, for use with the functions in this file.

    *Note*: Only uses the 'profile', 'imports', 'syspaths', 'mem', 'clean_files',
    'clean_outputs', and 'partition' keyword arguments for job submission,
    all others are ignored.

    Returns
    -------
    fyrd.job.Job
        A Job class (not submitted) for the command.
    """
    mem = kkk['mem'] if 'mem' in kkk else 2000
    cln_f = kkk['clean_files'] if 'clean_files' in kkk else True
    cln_o = kkk['clean_outputs'] if 'clean_outputs' in kkk else True
    kkk['clean_files'] = cln_f
    kkk['clean_outputs'] = cln_o
    add_args = dict(
        cores=1, mem=mem, clean_files=cln_f, clean_outputs=cln_o
    )
    if 'partition' in kkk:
        add_args.update(dict(partition=kkk['partition']))
    if 'profile' in kkk:
        add_args.update(dict(profile=kkk['profile']))
    if callable(command):
        _logme.log('Running function {} with export run'
                   .format(command.__name__), 'debug')
        kkk['imports']  = _run.export_imports(command, kkk)
        spth = [_run.get_function_path(command)]
        kkk['syspaths'] = spth + _run.listify(kkk['syspaths']) \
            if 'syspaths' in kkk else spth
        add_args.update(dict(imports=kkk['imports'], syspaths=kkk['syspaths']))
        _logme.log('export_run imports: {}'.format(kkk['imports']), 'debug')
        _logme.log('export_run syspath: {}'.format(kkk['syspaths']), 'debug')
    else:
        _logme.log('Running command {} with export run'
                   .format(command), 'debug')
    return _Job(
        _run.export_run, (command, args, kkk), **add_args
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Manage fyrd config, profiles, and queue.
"""

from __future__ import print_function
import os
import re
import sys
import argparse
import signal
from itertools import chain
from tabulate import tabulate

import fyrd

###############################################################################
#                                  Help Text                                  #
###############################################################################

DESC = """\
{doc}

============   ======================================
Author         Michael D Dacre <mike.dacre@gmail.com>
Organization   Stanford University
License        MIT License, use as you wish
Version        {version}
============   ======================================
""".format(doc=__doc__, version=fyrd.__version__)

RUN_HELP = """\
Run a shell script on the cluster and optionally wait for completion.

Allows the running of a single simple shell script, or the same shell script on
many files, or more complex file interpretation.
"""

RUN_EPI = """\
In simple mode (with `--simple` specified), all positional arguments are
converted into a single script and run on the cluster. Otherwise, the first
positional argument is the script and must be quoted. All other arguments
are file path specifiers.

When using file path speficiers, there are three possible modes, which will
be determined by the contents of the shell script.

1. Simple file addition:
    If the shell script contains '{}', all other positional arguments will be
    converted into a list, and the shell script will be submitted once for each
    file. e.g. ::

        fyrd run "my_algorithm {}" file1.txt my_dir/*.txt

    That will fun `my_algorithm` on every .txt file in my_dir plus on file1.txt
2. Complex file interpretation:
    Any '{#}' (e.g. {0}, {1}) in the shell script will be interpretted as a
    file, and the remaining positional arguments will be parsed to determine
    the files to use. e.g.::

        fyrd run "zcat {0} | my_algorithm --vcf {2} --stats {1}" \\
                 "my_dir/*.txt" "my_dir/*.stat" my_dir/info.vcf

    Note, in this mode the file blobs must interpret to be the same lengths,
    i.e. there must be as many .txt files in the directory as .stat files, and
    they must also sort identically (we will sort the lists alphabetically).
3. Complex file interpretation with variables:
    Exactly the same as two, but while '{#}' arguments will interpret to be
    the full path to a file (same as 2), it is possible to also include
    variables as '{name}'. These allow you to capture just one part of the
    path for use in the script. The exact same name should appear in the
    script itself to be used.

    Note: if you provide multiple file arguments that share a variable name,
    the string must be identical. For example, consider a directory with the
    files: {'sample_1.info.txt', 'sdata.sample_1.bat', 'sample_2.info.txt',
    'sdata.sample_2.bat'}. The following script would work::

        fyrd run "C=$(echo {sample} | sed 's/.*_//); \\
                  algorithm -n {sample} -c $C -b {1} {0} > {sample}.done" \\
                  "my_dir/{sample}.info.txt" "my_dir/sdata.{sample}.bat"

    In this case it would make more sense for the script to be a separate file,
    by putting the contents of the script into the file `my_script.sh`, the
    following command is equivalent to the last one::

        fyrd run my_script.sh "my_dir/{sample}.info.txt" \\
                              "my_dir/sdata.{sample}.bat"

Note: all of these 'modes' are actually compatible with each other and are
simply detected by parsing the arguments to fyrd. Only running a non-quoted
shell script with no file parsing requires an additional argument (`-s`).

If you want to use additional arguments, or to parse existing arguments,
you can pass the --extra-args argument, this allows you to define regular
expressions to alter arguments in the following format::

    name:orig_name:regex:sub,name:orig_name:regex:sub

`name` is what you would like the new variable to be called `orig_name` is the
original variable to parse, it can be the same as `name` if you want to edit
a variable. `regex` and `sub` are used to edit the contents of the variable.
For example, imagine a file name 'test_dir/sample4_parsed.bed.gz'::

    fyrd run --extra-vars "name:name:(.*)_.*:\\1" "parse_bed -o {name} {0}

The will result in the command::

    parse_bed -o sample4 /path/test_dir/sample4_parsed.bed.gz
"""

RUN_JOB_HELP = """\
Run an existing job file or set of files on the cluster and optionally wait
for completion.

e.g. fyrd run_file --wait ./jobs/*sh
"""

WAIT_HELP = """\
Wait on a list of jobs, block until they complete.
"""

CONF_HELP = """\
This script allows display and management of the fyrd config file found
here: {conf_file}.
""".format(conf_file=fyrd.conf.CONFIG_FILE)
CONF_EPI = """\
Show usage::
    fyrd conf show [-s <section>]

Update usage::
    fyrd conf update <section> <option> <value>

*Values can only be altered one at a time*

To create a new config from scratch interactively::
    fyrd conf init [--defaults]
"""
CONF_LIST_HELP = """\
By default shows the current config, which is based on the config file, but
with all parsing modifications applied, meaning that extra options or sections
are ignored.

To print the file as is, pass --file, or to limit to a single section, pass
--section <section>
"""

PROFILE_HELP = """\
Fyrd jobs use keyword arguments to run (for a complete list run this script
with the keywords command). These keywords can be bundled into profiles, which
are kept in {}. This file can be edited directly or manipulated here.
""".format(fyrd.conf.get_option('jobs', 'profile_file'))
PROF_EPI = """\
Show::
    fyrd prof show

Delete::
    fyrd prof delete <name>

Update::
    fyrd prof update <name> <options>

Add::
    fyrd prof add <name> <options>

<options>:
    The options arguments must be in the following format::
        opt:val opt2:val2 opt3:val3

Note: the DEFAULT profile is special and cannot be deleted, deleting it will
cause it to be instantly recreated with the default values. Values from this
profile will be available in EVERY other profile if they are not overriden
there. i.e. if DEFAULT contains `partition=normal`, if 'long' does not have
a 'partition' option, it will default to 'normal'.

To reset the profile to defaults, just delete the file and run this script
again.
"""

QUEUE_HELP = """\
Check the local queue, similar to squeue or qstat but simpler, good for
quickly checking the queue.

By default it searches only your own jobs, pass '--all-users' or
'--users <user> [<user2>...]' to change that behavior.

To just list jobs with some basic info, run with no arguments.
"""

DEFAULT_CONF_SECTIONS = set(fyrd.conf.DEFAULTS.keys())
DEFAULT_CONF_OPTS = set(
    chain(*[list(i.keys()) for i in fyrd.conf.DEFAULTS.values()])
)

with open(fyrd.conf.CONFIG_FILE) as fin:
    CURRENT_CONF = fin.read()

CONF_UPDATE = """\
Update one option at a time.

Sections: {}
""".format(DEFAULT_CONF_SECTIONS)
CONF_UPDATE_EPI = """\
Current config:
{}
""".format(CURRENT_CONF)

KEYWORD_INFO = """\

=================
fyrd keyword help
=================

All job submission functions in this module make use of the same simple
keyword arguments. These keywords can be specified directly in `Job` based
function calls::

    fyrd.submit('ls', cores=10, mem=8000)

They can also be bundled together into profiles with the cluster-profile
script, and then specified as a profile::

    fyrd.submit('ls', profile='large')

Available keywords
==================

"""

CLEAN_HELP = """\
Clean all intermediate files created by the cluster module.

If not directory is passed, the default if either scriptpath or outpath are
set in the config is to clean files in those locations is to clean those
directories. If they are not set, the default is the current directory.

By default, outputs are not cleaned, to clean them too, pass '-o'

Caution:
    The clean() function will delete **EVERY** file with
    extensions matching those these::

        .<suffix>.err
        .<suffix>.out
        .<suffix>.sbatch & .fyrd.script for slurm mode
        .<suffix>.qsub for torque mode
        .<suffix> for local mode
        _func.<suffix>.py
        _func.<suffix>.py.pickle.in
        _func.<suffix>.py.pickle.out

"""

###############################################################################
#                         Catch Keyboard Interruption                         #
###############################################################################

def catch_keyboard(sig, frame):
    """Catch Keyboard Interruption."""
    sys.stderr.write('\nKeyboard Interrupt Detected, Exiting\n')
    sys.exit(1)

signal.signal(signal.SIGINT, catch_keyboard)

###############################################################################
#                Functions to work with Command Line Arguments                #
###############################################################################


############
#  Config  #
############

def config(args):
    """Handle config management."""
    if not args.cmnd:
        sys.stdout.write('Subcommands required, for help run fyrd conf -h\n')
        return 1
    print(args)


def show_config(args):
    """Print the current config."""
    if args.file:
        print('Current config file ({}):\n'.format(fyrd.conf.CONFIG_FILE))
        with open(fyrd.conf.CONFIG_FILE) as fin1:
            print(fin1.read())
    else:
        print('Current config, the parsed output of the config file:\n')
        sections = args.sections if args.sections \
                                else fyrd.conf.DEFAULTS.keys()
        sections = sorted(sections, key=_sort_conf)
        for section in sections:
            print('[{}]'.format(section))
            sect = fyrd.conf.get_option(section=section)
            for opt, val in sect.items():
                print('{}: {}'.format(opt, val))
            print()
    print('For an explanation of options, run fyrd conf help')


def display_conf_help(args):
    """Print helpful information about the config options."""
    print(fyrd.conf.CONF_HELP['summary'])
    sections = args.sections if args.sections \
                            else fyrd.conf.DEFAULTS.keys()
    sections = sorted(sections, key=_sort_conf)
    for section in sections:
        print(fyrd.conf.CONF_HELP[section])


def update_option(args):
    """Update a single config option."""
    fyrd.conf.set_option(args.section, args.option, args.value)
    print('Done')


def init_config(args):
    """Start config initialization."""
    if args.defaults:
        if args.yes:
            ans = fyrd.run.get_input('Overwrite current config? [y/N] ')
            if ans.lower() != 'y':
                print('Aborting')
                return 1
            fyrd.conf.create_config()
    else:
        fyrd.conf.create_config_interactive(prompt=args.yes)
    print('Done\n')
    print('New config:\n')
    for section in sorted(fyrd.conf.DEFAULTS, key=_sort_conf):
        print('[{}]'.format(section))
        sect = fyrd.conf.get_option(section=section)
        for opt, val in sect.items():
            print('{}: {}'.format(opt, val))
        print()
    print('Current profiles:\n')
    list_profiles()


#############
#  Profile  #
#############


def profile(args):
    """This shouldn't be called."""
    print(args)
    sys.stdout.write('Subcommands required, for help run fyrd prof -h\n')
    return 1


def list_profiles(_=None):
    """List all profiles."""
    profiles = fyrd.conf.get_profiles()
    print(profiles.pop('DEFAULT'))
    for prof in profiles.values():
        print(prof)


def add_profile(args):
    """Add a profile."""
    add_edit_profile(args, False)


def edit_profile(args):
    """Edit a profile."""
    add_edit_profile(args, True)


def del_profile(args):
    """Delete a profile."""
    args.name = 'DEFAULT' if args.name.lower() == 'default' else args.name

    print('This will delete the {} profile.'.format(args.name))
    if fyrd.run.get_input('Are you sure? [y/N] ', ['y', 'n']).lower() == 'y':
        fyrd.conf.del_profile(args.name)
        print('Done')
    else:
        print('Aborting')

    if args.name == 'DEFAULT':
        sys.stderr.write('Cannot delete DEFAULT profile, ' +
                         'resetting it instead\n')
        fyrd.conf.load_profiles()


def delete_profile_option(args):
    """Remove an option from a profile."""
    for opt in args.options:
        print('Removing {} from {}'.format(opt, args.section))
        fyrd.conf.profiles.remove_option(args.section, opt)
    print('Done')

###################
#  Running stuff  #
###################


def run_args_to_keywords(args):
    """Parse the results of the run_modes parser to create keyword args.

    Parameters
    ----------
    args : Namespace
        Argparse command line arguments defined in main.

    Returns
    -------
    fyrd_keywords : dict
    """
    if not args.wait and args.clean:
        raise ValueError('--clean requires --wait, use `fyrd clean` to clean '
                         'without waiting')
    kwargs = {}
    if args.args:
        for arg in args.args.split(','):
            sarg = arg.split('=')
            if len(sarg) == 1:
                kwargs[sarg[0]] = None
            elif len(sarg) == 2:
                kwargs[sarg[0]] = sarg[1]
            else:
                raise TypeError(
                    'Invalid argument: {arg}, must be key=value, or just key'
                )
    if args.cores:
        kwargs['cores'] = args.cores
    if args.mem:
        kwargs['memory'] = args.mem
    if args.time:
        kwargs['time'] = args.time
    kwargs['clean_files'] = args.keep
    kwargs['clean_outputs'] = args.clean
    return kwargs


def run(args):
    """Run an arbitrary shell script as a job."""
    #  r = re.compile(r'{(.*?)}')
    jobs = []
    kwargs = run_args_to_keywords(args)
    extra_vars = args.extra_vars.split(',') if args.extra_vars else []

    if args.simple:
        if args.file_parsing:
            command = ' '.join([args.shell_script] + args.file_parsing)
        else:
            command = args.shell_script
        command = fyrd.run.cmd_or_file(command)
        job = fyrd.job.Job(command, profile=args.profile, **kwargs).submit()
        print('Job submitted as job {0}'.format(job.id))
        jobs.append(job)
    else:
        if not args.shell_script:
            sys.stderr.write('Shell script is required in run mode')
            return 1
        command = fyrd.run.cmd_or_file(args.shell_script)
        script_ints, script_vars = fyrd.run.string_getter(command)
        for fl in args.file_parsing:
            _, fvars = fyrd.run.string_getter(fl)
            script_vars.update(fvars)
        for evar in extra_vars:
            new_var, orig_var = evar.split(':')[:2]
            if orig_var not in script_vars:
                raise ValueError('Original variable from {} not in script or '
                                 .format(evar) + 'file parsing strings')
            script_vars.update({new_var, orig_var})
        # Simple file mode
        if '{}' in args.shell_script:
            if not args.file_parsing:
                sys.stderr.write("Files must be provided if '{}' is in the "
                                 "command and `--simple` mode is off")
                return 2
            # Just in case someone quoted the file glob
            if len(args.file_parsing) == 1:
                files = list(fyrd.run.parse_glob(args.file_parsing[0]).keys())
            else:
                files = args.file_parsing
            for fl in files:
                command = command.format(fl)
                if args.dry_run:
                    print(command)
                else:
                    job = fyrd.job.Job(command, profile=args.profile, **kwargs)
                    job = job.submit()
                    print('{0} ({1}): {2}'.format(job.id, job.uuid, fl))
                    jobs.append(job)
        else:
            file_info = fyrd.run.file_getter(
                args.file_parsing, script_vars,
                extra_vars=extra_vars, max_count=len(script_ints)
            )
            for fls, fvars in file_info:
                command = command.format(*fls, **fvars)
                if args.dry_run:
                    print(command)
                else:
                    job = fyrd.job.Job(command, profile=args.profile, **kwargs)
                    job = job.submit()
                    print('{0} ({1}): {2}'.format(job.id, job.uuid, fls))
                    jobs.append(job)

    if args.wait and not args.dry_run:
        print('Waiting for job(s) to complete')
        if args.notify:
            arg = args.email if args.email else True
            fyrd.wait(jobs, notify=arg)
        elif args.no_notify:
            q.wait(args.jobs, notify=False)
        else:
            fyrd.wait(jobs)


def sub_files(args):
    """Run any number of existing job files and optionally wait for them."""
    for job in args.job_files:
        if not os.path.isfile(job):
            sys.stderr.write('Job file {0} does not exist'.format(job))
            sys.exit(1)
    job_nos = []
    for job in args.job_files:
        job_no = fyrd.basic.submit_file(job)
        print('{0}: {1}'.format(job_no, job))
        job_nos.append(job_no)
    if args.wait:
        print('Waiting for job to complete')
        if args.notify:
            arg = args.email if args.email else True
            fyrd.wait(job_nos, notify=arg)
        elif args.no_notify:
            q.wait(args.jobs, notify=False)
        else:
            fyrd.wait(job_nos)


#######################
#  Other Subcommands  #
#######################


def keyword_help(args):
    """Print keyword info."""
    if args.split_tables:
        print(KEYWORD_INFO + fyrd.options.option_help(mode='table'))
    elif args.table:
        print(KEYWORD_INFO + fyrd.options.option_help(mode='merged_table'))
    elif args.list:
        print(fyrd.options.option_help(mode='list'))
    else:
        print(KEYWORD_INFO + fyrd.options.option_help())


def queue(args):
    """Handle queue management."""
    # Create queue object
    if args.all_users or args.users:
        q = fyrd.queue.Queue()
    else:
        q = fyrd.queue.Queue(user='self')

    # Get jobs
    if args.running:
        jobs = q.running
    elif args.queued:
        jobs = q.queued
    elif args.done:
        jobs = q.completed
    elif args.bad:
        jobs = q.bad
    else:
        jobs = q.jobs

    # Filter users
    if args.users:
        jobs = {i: j for i, j in jobs.items() if j.owner in args.users}

    if not jobs:
        jobs = {}

    # Print requested output
    if args.count:
        print("{0}".format(len(jobs)))
    elif args.list:
        print("{0}".format('\n'.join([str(i) for i in jobs.keys()])))
    else:
        out_table = []
        arr = False
        for jid, job in jobs.items():
            cols = [jid, job.name, job.state]
            if job.array_job:
                arr = True
                cols.append(str(len(job.children)))
            if args.users or args.all_users:
                cols.append(job.owner)
            if args.nodes:
                nodes = {}
                for node in job.nodes:
                    node = node.split('/')[0]
                    if node in nodes:
                        nodes[node] += 1
                    else:
                        nodes[node] = 1
                cols.append(','.join(['{0}({1})'.format(n,c)
                                      for n, c in nodes.items()]))
            out_table.append(cols)
        headers = ['ID', 'Name', 'State']
        if args.users or args.all_users:
            headers.append('User')
        if arr:
            headers.append('Array Job Count')
        if args.nodes:
            headers.append('Node(s)')
        print('\n' + tabulate(out_table, headers=headers) + '\n')


def wait(args):
    """Wait on jobs."""
    q = fyrd.queue.Queue()
    if args.users:
        users = args.users.strip().split(',')
        args.jobs += list(q.get_user_jobs(users).values())
    if args.notify:
        arg = args.email if args.email else True
        q.wait(args.jobs, notify=arg)
    elif args.no_notify:
        q.wait(args.jobs, notify=False)
    else:
        q.wait(args.jobs)


def clean_dir(args):
    """Clean up a job directory."""
    run_tmp_clean = False
    if args.dir:
        directory = args.dir
    else:
        scriptpath = fyrd.conf.get_option('jobs', 'scriptpath')
        outpath    = fyrd.conf.get_option('jobs', 'outpath')
        if scriptpath:
            run_tmp_clean = True
        if outpath and args.outputs:
            run_tmp_clean = True
        if not run_tmp_clean:
            directory = os.path.abspath('.')

    if run_tmp_clean:
        files = fyrd.basic.clean_work_dirs(outputs=args.outputs,
                                           confirm=args.no_confirm)
    else:
        files = fyrd.basic.clean_dir(directory=directory, suffix=args.suffix,
                                     qtype=args.qtype, confirm=args.no_confirm,
                                     delete_outputs=args.outputs)

    if not files:
        print('No files deleted.')

    # Print list of files if it wasn't done by the function
    elif not args.no_confirm:
        sys.stdout.write('Deleted files:\n\t')
        sys.stdout.write('\n\t'.join(files))
        sys.stdout.write('\n')

############################
#  Local Queue Management  #
############################


def manage_daemon(args):
    """Start, stop, or restart the local queue daemon."""
    from fyrd.batch_systems import local
    return local.daemon_manager(args.mode)


######################
#  Helper Functions  #
######################


def add_edit_profile(args, overwrite):
    """Add or edit a profile.

    Parameters
    ----------
    args : Namespace
        Command line arguments from argparse
    overwrite : bool
        Edit a profile instead of adding one.
    """
    args.name = 'DEFAULT' if args.name.lower() == 'default' else args.name
    # Parse values
    values = get_values(args.options)

    if not overwrite and fyrd.conf.get_profile(args.name):
        sys.stderr.write("Profile '{}' already exists, use edit to change.\n"
                         .format(args.name))
        sys.exit(2)

    # Try to set the profile
    try:
        fyrd.conf.set_profile(args.name, values)
        print('Success')
    except fyrd.options.OptionsError as err:
        sys.stderr.write(err + '\n')
        sys.exit(3)


def get_values(keywords):
    """Return a dictionary of {keyword: arg} from a list of 'keyword:arg'.

    Parameters
    ----------
    keywords : list
        A list of strings in the format keyword:arg.

    Returns
    -------
    dict
        A dictionary of keyword arguments.
    """
    values = {}
    for arg in keywords:
        try:
            key, value = arg.split(':', 1)
        except ValueError:
            sys.stderr.write('Arguments must be in the format keyword:arg\n')
            sys.exit(1)
        values[key] = value
    return values


def _sort_conf(key):
    """Use with sorted: return an integer value for each section in config."""
    sections = {'queue': 1, 'jobs': 2, 'jobqueue': 3}
    return sections[key]


###############################################################################
#                            Command Line Parsing                             #
###############################################################################

class AliasedSubParsersAction(argparse._SubParsersAction):

    """Allow aliased subparser arguments in python2.

    From: https://gist.github.com/sampsyo/471779

    """

    class _AliasedPseudoAction(argparse.Action):

        """Print alias info."""

        def __init__(self, name, aliases, hlp):
            """Print alias info."""
            dest = name
            if aliases:
                dest += ' (%s)' % ','.join(aliases)
            sup = super(AliasedSubParsersAction._AliasedPseudoAction, self)
            sup.__init__(option_strings=[], dest=dest, help=hlp)

    def add_parser(self, name, **kwargs):
        """Add a parser with aliases."""
        if 'aliases' in kwargs:
            aliases = kwargs['aliases']
            del kwargs['aliases']
        else:
            aliases = []

        assert 'aliases' not in kwargs
        parser = super(AliasedSubParsersAction, self).add_parser(
            name, **kwargs)

        # Make the aliases work.
        for alias in aliases:
            self._name_parser_map[alias] = parser
        # Make the help text reflect them, first removing old help entry.
        if 'help' in kwargs:
            hlp = kwargs.pop('help')
            self._choices_actions.pop()
            pseudo_action = self._AliasedPseudoAction(name, aliases, hlp)
            self._choices_actions.append(pseudo_action)

        return parser

def command_line_parser():
    """Parse command line options.

    Returns:
        argparse parser
    """
    parser = argparse.ArgumentParser(
        description=DESC,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    if sys.version_info.major == 2:
        parser.register('action', 'parsers', AliasedSubParsersAction)

    # Global arguments
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Show debug outputs')
    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s {0}'.format(fyrd.__version__),
                        help='Print version string')

    # Subcommands
    modes = parser.add_subparsers(
        dest='modes',
        metavar='{run,submit,wait,queue,conf,prof,keywords,clean,local}'
    )

    #################################
    #  Parent Parser for Run Modes  #
    #################################

    run_modes = argparse.ArgumentParser(add_help=False)
    run_opts = run_modes.add_argument_group("Run Options")
    run_opts.add_argument('-p', '--profile',
                          help='The profile to use to run')
    run_opts.add_argument('-c', '--cores', type=int,
                          help='The number of cores to request')
    run_opts.add_argument('-m', '--mem',
                          help='The amount of memory to request')
    run_opts.add_argument('-t', '--time',
                          help='The amount of walltime to request')
    run_opts.add_argument('-a', '--args',
                          help='Submission args, e.g.: ' +
                          "'time=00:20:00,mem=20G,cores=10'")
    run_opts.add_argument('-w', '--wait', action='store_true',
                          help='Wait for the job to complete')
    run_opts.add_argument('-k', '--keep', action='store_false',
                          help='Keep submission scripts')
    run_opts.add_argument('-l', '--clean', action='store_true',
                          help='Delete STDOUT and STDERR files when done')

    ##################################
    #  Waiting Notification Options  #
    ##################################

    default_email = fyrd.conf.get_option('notify', 'notify_address')
    wait_modes = argparse.ArgumentParser(add_help=False)
    wait_opts = wait_modes.add_argument_group("Notification Options")
    wait_opts.add_argument('-n', '--notify', action='store_true',
                           help='Send notification email when done')
    wait_opts.add_argument('--no-notify', action='store_true',
                           help='Suppress automatic notifications')
    wait_opts.add_argument('-e', '--email', default=default_email,
                           help=('Email address to send notification to, '
                                 'default set in ~/.fyrd/config.txt'))

    ######################
    #  Run Shell Script  #
    ######################

    run_sub = modes.add_parser(
        'run', description=RUN_HELP, epilog=RUN_EPI,
        parents=[run_modes, wait_modes], help="Run simple shell scripts",
        aliases=['r'],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    run_sub.add_argument('-s', '--simple', action='store_true',
                         help='The amount of walltime to request')
    run_sub.add_argument('-x', '--extra-vars',
                         help='Regex in form "new_var:orig_var:regex:sub,..."')
    run_sub.add_argument('-d', '--dry-run', action='store_true',
                         help='Print commands instead of running them')
    run_sub.add_argument('shell_script', nargs='?',
                         help="The script to run")
    run_sub.add_argument('file_parsing', nargs='*',
                         help="The script to run")

    # Set function
    run_sub.set_defaults(func=run)

    ########################
    #  Run Existing Files  #
    ########################

    run_job_sub = modes.add_parser(
        'submit', description=RUN_HELP, help="Submit existing job files",
        parents=[run_modes, wait_modes], aliases=['sub', 's'],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    run_job_sub.add_argument('job_files', nargs='+',
                             help="The script to run")

    # Set function
    run_job_sub.set_defaults(func=sub_files)

    #################
    #  Job Waiting  #
    #################

    wait_sub = modes.add_parser(
        'wait', description=WAIT_HELP, help="Wait for jobs",
        parents=[wait_modes], aliases=['w'],
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # We use * here instead of + because it is possible to wait on all jobs
    # for a user, making the jobs argument unnecessary.
    wait_sub.add_argument('jobs', nargs='*', help="Job list to wait for")
    wait_sub.add_argument('-u', '--users',
                          help='A comma-separated list of users to wait for')

    # Set function
    wait_sub.set_defaults(func=wait)

    ###################
    #  Queue Parsing  #
    ###################

    queue_sub = modes.add_parser(
        'queue', aliases=['q'],
        description=QUEUE_HELP, help="Search the queue",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # User and partition filtering
    queue_filter = queue_sub.add_argument_group('queue filtering')
    queue_filter_m = queue_filter.add_mutually_exclusive_group()
    queue_filter_m.add_argument('-u', '--users', nargs='+', metavar='',
                                help='Limit to these users')
    queue_filter_m.add_argument('-a', '--all-users', action='store_true',
                                help='Display jobs for all users')
    queue_filter.add_argument('-p', '--partitions', nargs='+', metavar='',
                              help="Limit to these partitions (queues)")

    # State filtering
    queue_filter = queue_sub.add_argument_group('queue state filtering')
    queue_filter_s = queue_filter.add_mutually_exclusive_group()
    queue_filter_s.add_argument('-r', '--running', action='store_true',
                                help="Show only running jobs")
    queue_filter_s.add_argument('-q', '--queued', action='store_true',
                                help="Show only queued jobs")
    queue_filter_s.add_argument('-d', '--done', action='store_true',
                                help="Show only completed jobs")
    queue_filter_s.add_argument('-b', '--bad', action='store_true',
                                help="Show only completed jobs")

    # Display mode
    queue_disp_group = queue_sub.add_argument_group('display options')
    queue_disp = queue_disp_group.add_mutually_exclusive_group()
    queue_disp.add_argument('-l', '--list', action='store_true',
                            help="Print job numbers only, works well with " +
                            "xargs")
    queue_disp.add_argument('-c', '--count', action='store_true',
                            help="Print job count only")
    queue_disp.add_argument('-n', '--nodes', action='store_true',
                            help='Also display node list (normal mode only)')

    # Set function
    queue_sub.set_defaults(func=queue)

    #########################
    #  Config Manipulation  #
    #########################

    conf_sub = modes.add_parser(
        'conf', description=CONF_HELP, epilog=CONF_EPI, aliases=['config'],
        help="View and manage the config",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    if sys.version_info.major == 2:
        conf_sub.register('action', 'parsers', AliasedSubParsersAction)
    conf = conf_sub.add_subparsers(dest='cmnd')

    # Printing config
    conf_list = conf.add_parser('show', aliases=['list'],
                                description=CONF_LIST_HELP,
                                help="Show current config")
    conf_list_args = conf_list.add_mutually_exclusive_group()
    conf_list.add_argument('-s', '--sections', metavar='', nargs='+',
                           choices=DEFAULT_CONF_SECTIONS,
                           help="Limit results to a list of sections")
    conf_list_args.add_argument('-f', '--file', action='store_true',
                                help="Print file contents only")
    conf_list.set_defaults(func=show_config)

    # Config help
    conf_help = conf.add_parser('help',
                                help="Show info on every config option")
    conf_help.add_argument('-s', '--sections', metavar='', nargs='+',
                           choices=DEFAULT_CONF_SECTIONS,
                           help="Limit results to a list of sections")
    conf_help.set_defaults(func=display_conf_help)

    # Updating config
    conf_update = conf.add_parser(
        'update', help="Update the config", aliases=['alter'],
        description=CONF_UPDATE, epilog=CONF_UPDATE_EPI,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    conf_update.add_argument('section', help="Section to update",
                             choices=DEFAULT_CONF_SECTIONS,
                             metavar='section')
    conf_update.add_argument('option', help="Option to update",
                             choices=DEFAULT_CONF_OPTS, metavar='option')
    conf_update.add_argument('value', help="New value for option")
    conf_update.set_defaults(func=update_option)

    # Initializing config
    conf_init = conf.add_parser(
        'init', help="Interactively initialize the config",
        description="Overwrites the exiting config and starts from scratch. " +
        "Asks for input from user for default options."
    )
    conf_init.add_argument('--defaults', action='store_true',
                           help="Non-interactive, just use builtin defaults")
    conf_init.add_argument('--yes', action='store_false',
                           help="Do not as for confirmation")
    conf_init.set_defaults(func=init_config)

    # Set function
    conf_sub.set_defaults(func=config)

    ######################
    #  Profile Handling  #
    ######################

    prof_sub = modes.add_parser(
        'prof', description=PROFILE_HELP, epilog=PROF_EPI, aliases=['profile'],
        help="Manage profiles",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    if sys.version_info.major == 2:
        prof_sub.register('action', 'parsers', AliasedSubParsersAction)
    prof = prof_sub.add_subparsers(dest='cmnd')

    # List profiles
    prof_show = prof.add_parser('show', aliases=['list'],
                                help="Print current profiles")
    prof_show.add_argument('-n', '--name', nargs='+',
                           help="Limit to only these profiles")
    prof_show.set_defaults(func=list_profiles)

    # Add a profile
    prof_add = prof.add_parser('add', aliases=['new'],
                               help="Add a new profile")
    prof_add.add_argument('name', help='Profile name')
    prof_add.add_argument('options', nargs='+', help="Options to update")
    prof_add.set_defaults(func=add_profile)

    # Update profile
    prof_update = prof.add_parser('update', aliases=['alter', 'edit'],
                                  help="Update an existing profile")
    prof_update.add_argument('name', help='Profile name')
    prof_update.add_argument('options', nargs='+', help="Options to update")
    prof_update.set_defaults(func=edit_profile)

    # Remove an option
    prof_optdel = prof.add_parser('remove-option', aliases=['del-option'],
                                  help="Remove a profile option")
    prof_optdel.add_argument('name', help='Profile name')
    prof_optdel.add_argument('options', nargs='+', help="Options to remove")
    prof_optdel.set_defaults(func=delete_profile_option)

    # Delete a profile
    prof_del = prof.add_parser('delete', aliases=['del'],
                               help="Delete an existing profile")
    prof_del.add_argument('name', help='Profile name')
    prof_del.set_defaults(func=del_profile)

    ##################
    #  Keyword Help  #
    ##################

    keywords = modes.add_parser('keywords', aliases=['keys', 'options'],
                                help="Print available keyword arguments.")
    keywords_grp = keywords.add_mutually_exclusive_group()
    keywords_grp.add_argument('-t', '--table', action='store_true',
                              help="Print keywords as a table")
    keywords_grp.add_argument('-s', '--split-tables', action='store_true',
                              help="Print keywords as multiple tables")
    keywords_grp.add_argument('-l', '--list', action='store_true',
                              help="Print a list of keywords only")
    keywords.set_defaults(func=keyword_help)

    ########################
    #  Directory Cleaning  #
    ########################

    clean = modes.add_parser(
        'clean', description=CLEAN_HELP, help="Clean up a job directory",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    clean.add_argument('dir', nargs='?',
                       help="Directory to clean (optional)")

    clean.add_argument('-o', '--outputs', action='store_true',
                       help="Clean output files too")
    clean.add_argument('-s', '--suffix',
                       default=fyrd.conf.get_option('jobs', 'suffix'),
                       help="Suffix to use for cleaning")
    clean.add_argument('-q', '--qtype', choices=('torque', 'slurm', 'local'),
                       help="Limit deletions to this qtype")

    # We store this as false as the question is a negative
    clean.add_argument('-n', '--no-confirm', action='store_false',
                       help="Do not confirm before deleting (for scripts)")

    # Set function
    clean.set_defaults(func=clean_dir)

    ############################
    #  Local Queue Management  #
    ############################

    server_mode = modes.add_parser(
        'local', aliases=['server'], help='Manage the local queue server'
    )
    server_mode.add_argument(
        'mode', choices={'start', 'stop', 'status', 'restart'},
        metavar='{start,stop,status,restart}', help='Server command')

    # Set function
    server_mode.set_defaults(func=manage_daemon)

    return parser


###############################################################################
#                             Running as a Script                             #
###############################################################################


def main(argv=None):
    """Parse command line options to run as a script."""
    if not argv:
        argv = sys.argv[1:]

    parser = command_line_parser()

    args = parser.parse_args(argv)

    if not args.modes:
        parser.print_help()
        return 0

    if args.verbose:
        fyrd.logme.MIN_LEVEL = 'debug'

    # Call the subparser function
    try:
        return args.func(args)
    except TypeError:
        print(args)
        raise


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
"""
Monitor the queue for torque or slurm.

Provides a class to monitor the all defined batch queues with identical
syntax.

At its simplest, you can use it like::

    q = queue.Queue()
    q.jobs
    q.running
    q.pending
    q.complete

All of the above commands return a dictionary of:
    {job_no: Queue.QueueJob}

Queue.QueueJob classes include information on job state, owner, queue, nodes,
threads, exitcode, etc.

Queue also defines a wait() method that takes a list of job numbers, job.Job()
objects, or JobQueue.Job objects and blocks until those jobs to complete

The default cluster environment is also defined in this file as MODE, it can be
set directly or with the get_cluster_environment() function definied here.
"""
import sys as _sys          # Used to get TB info
import pwd as _pwd          # Used to get usernames for queue
import getpass as _getpass  # Used to get usernames for queue
import traceback as _tb     # Used to mail TB info
from datetime import datetime as _dt
from time import time as _time
from time import sleep as _sleep

from six import reraise as _reraise
from six import text_type as _txt
from six import string_types as _str
from six import integer_types as _int

###############################################################################
#                                Our functions                                #
###############################################################################

from . import run as _run
from . import logme as _logme
from . import conf as _conf
from . import ClusterError as _ClusterError
from . import batch_systems as _batch
from . import notify as _notify

# Funtions to import if requested
__all__ = ['Queue']

GOOD_STATES      = _batch.GOOD_STATES
ACTIVE_STATES    = _batch.ACTIVE_STATES
BAD_STATES       = _batch.BAD_STATES
UNCERTAIN_STATES = _batch.UNCERTAIN_STATES
ALL_STATES       = _batch.ALL_STATES
DONE_STATES      = _batch.DONE_STATES

###############################################################################
#                               The Queue Class                               #
###############################################################################


class Queue(object):

    """A wrapper for all defined batch systems.

    Attributes
    ----------
    jobs : dict
        A dictionary of all jobs in this queue in the form:
        `{jobid: Queue.QueueJob}`
    finished : dict
        A dictionary of all completed jobs, same format as jobs
    bad : dict
        A dictionary of all jobs with failed or unknown states, same format as
        jobs
    active_job_count : int
        Total jobs in the queue (including array job children)
    max_jobs : int
        The maximum number of jobs allowed in the queue
    can_submit : bool
        True if active_job_count < max_jobs, False otherwise
    job_states : list
        A list of the different states of jobs in this queue
    active_job_count : int
        A count of all jobs that are either pending or running in the current
        queue
    can_submit : bool
        True if total active jobs is less than max_jobs
    users : set
        A set of all users with active jobs
    job_states : set
        A set of all current job states

    Methods
    -------
    wait(jobs, return_disp=False)
        Block until all jobs in jobs are complete.
    get(jobs)
        Get all results from a bunch of Job objects.
    wait_to_submit(max_jobs=None)
        Block until fewer running/pending jobs in queue than max_jobs.
    update()
        Refresh the list of jobs from the server.
    get_jobs(key)
        Return a dict of jobs where state matches key.
    get_user_jobs(users)
        Return a dict of jobs for all all jobs by each user in users.
    """

    def __init__(self, user=None, partition=None, qtype=None):
        """Can filter by user, queue type or partition on initialization.

        Parameters
        ----------
        user : str
            Optional usernameto filter the queue with.  If user='self' or
            'current', the current user will be used.
        partition : str
            Optional partition to filter the queue with.
        qtype : str
            one of the defined batch queues (e.g. 'slurm')
        """
        # Get user ID as an int UID
        if user:
            if user == 'self' or user == 'current':
                self.user = _getpass.getuser()
                """The username if defined."""
                self.uid  = _pwd.getpwnam(self.user).pw_uid
            elif user == 'ALL':
                self.user = None
            else:
                if isinstance(user, int) \
                        or (isinstance(user, (_str, _txt)) and user.isdigit()):
                    self.uid  = int(user)
                else:
                    self.uid = _pwd.getpwnam(str(user)).pw_uid
        else:
            self.uid = None
        self.user = _pwd.getpwuid(self.uid).pw_name if self.uid else None
        self.partition = partition
        self.max_jobs = int(_conf.get_option('queue', 'max_jobs'))
        # Don't allow max jobs to be less than 5, otherwise basic split jobs
        # permanently hang
        if self.max_jobs < 4:
            self.max_jobs = 4

        # Support python2, which hates reciprocal import
        from .job import Job
        self._Job      = Job

        # Get sleep time and update time
        self.queue_update_time = float(
            _conf.get_option('queue', 'queue_update', 2)
        )
        self.sleep_len = float(_conf.get_option('queue', 'sleep_len', 0.5))

        # Set type
        if qtype:
            _batch.check_queue(qtype)
        else:
            _batch.check_queue()
        self.qtype = qtype if qtype else _batch.MODE
        self.batch_system = _batch.get_batch_system(self.qtype)

        # Allow tracking of updates to prevent too many updates
        self._updating = False

        # Will contain a dict of QueueJob objects indexed by ID
        self.jobs = {}
        """All jobs currently in this queue."""

        self._update()

    ####################
    #  Public Methods  #
    ####################

    def check_dependencies(self, dependencies):
        """Check if dependencies are running.

        Parameters
        ----------
        dependencies : list
            List of job IDs

        Returns
        -------
        str
            'active' if dependencies are running or queued, 'good' if
            completed, 'bad' if failed, cancelled, or suspended, 'absent'
            otherwise.
        """
        for dep in _run.listify(dependencies):
            dep = str(dep)
            if dep not in self.jobs:
                return 'absent'
            state = self.jobs[dep].state
            if state in ACTIVE_STATES:
                return 'active'
            elif state in GOOD_STATES:
                continue
            elif state in BAD_STATES or state in UNCERTAIN_STATES:
                return 'bad'
            else:
                raise _ClusterError('Invalid state {0}'.format(state))
        return 'good'

    def wait(self, jobs, return_disp=False, notify=True):
        """Block until all jobs in jobs are complete.

        Update time is dependant upon the queue_update parameter in your
        ~/.fyrd/config.txt file.

        Parameters
        ----------
        jobs : list
            List of either fyrd.job.Job, fyrd.queue.QueueJob, job_id
        return_disp : bool, optional
            If a job disappeares from the queue, return 'disapeared' instead of
            True
        notify : str, True, or False, optional
            If True, both notification address and wait_time must be set in
            the [notify] section of the config. A notification email will be
            sent if the time exceeds this time. This is the default.
            If a string is passed, notification is forced and the string must
            be the to address.
            False means no notification

        Returns
        -------
        bool or str
            True on success False or None on failure unless return_disp is True
            and the job disappeares, then returns 'disappeared'
        """
        if notify is True:
            notify_time = _conf.get_option('notify', 'wait_time', 120)
            notify_addr = _conf.get_option('notify', 'notify_address')
        elif isinstance(notify, str):
            notify_addr = notify
            notify_time = 0
        start_time  = _dt.now()
        self.update()
        _logme.log('Queue waiting.', 'debug')

        # Sanitize arguments
        jobs = _run.listify(jobs)
        for job in jobs:
            if not isinstance(job, (_str, _txt, _int, QueueJob, self._Job)):
                raise _ClusterError('job must be int, string, or Job, ' +
                                    'is {}'.format(type(job)))

        check_jobs = []
        for job in jobs:
            if isinstance(job, (self._Job, QueueJob)):
                job_id = job.id
            else:
                job_id = str(job)
            check_jobs.append(job_id)

        pbar = _run.get_pbar(jobs, name="Waiting for job completion",
                             unit='jobs')
        dispo = 'Unknown'
        msg = None
        try:
            while check_jobs:
                self.update()
                for job in check_jobs:
                    if isinstance(job, (self._Job, QueueJob)):
                        job_id = job.id
                    else:
                        job_id = str(job)
                    job_id, array_id = self.batch_system.normalize_job_id(job_id)
                    _logme.log('Checking {}'.format(job_id), 'debug')
                    lgd = False
                    # Allow 12 seconds to elapse before job is found in queue,
                    # if it is not in the queue by then, assume failure.
                    if not self.test_job_in_queue(job_id):
                        if return_disp:
                            dispo = 'disappeared'
                            break
                        else:
                            dispo = False
                            break
                    ## Actually look for job in running/queued queues
                    lgd      = False
                    lgd2     = False
                    start    = _dt.now()
                    res_time = float(_conf.get_option('queue', 'res_time'))
                    count    = 0
                    # Get job state
                    job_state = self.jobs[job_id].state
                    # Check the state
                    if job_state in GOOD_STATES:
                        _logme.log('Queue wait for {} complete'
                                   .format(job_id), 'debug')
                        check_jobs.pop(check_jobs.index(job))
                        pbar.update()
                        dispo = True
                        break
                    elif job_state in ACTIVE_STATES:
                        if lgd:
                            _logme.log('{} not complete yet, waiting'
                                       .format(job_id), 'debug')
                            lgd = True
                        else:
                            _logme.log('{} still not complete, waiting'
                                       .format(job_id), 'verbose')
                    elif job_state in BAD_STATES:
                        msg = 'Job {} failed with state {}'.format(job, job_state)
                        _logme.log(msg, 'error')
                        dispo = False
                        break
                    elif job_state in UNCERTAIN_STATES:
                        if not lgd2:
                            _logme.log('Job {} in state {}, waiting {} '
                                       .format(job, job_state, res_time) +
                                       'seconds for resolution', 'warn')
                            lgd2 = True
                        if (_dt.now() - start).seconds > res_time:
                            msg = (
                                'Job {} still in state {} after {}s wait, aborting'
                                .format(job, job_state, res_time)
                            )
                            _logme.log(msg, 'error')
                            dispo = False
                            break
                    else:
                        if count == 5:
                            _logme.log('Job {} in unknown state {} '
                                       .format(job, job_state) +
                                       'cannot continue', 'critical')
                            raise QueueError('Unknown job state {}'
                                             .format(job_state))
                        _logme.log('Job {} in unknown state {} '
                                   .format(job, job_state) +
                                   'trying to resolve', 'debug')
                        count += 1
                if dispo in [False, 'disappeared']:
                    break
                _sleep(self.sleep_len)
            # Update jobs
            for job in jobs:
                if isinstance(job, self._Job):
                    job.update()
        except:
            dispo = _sys.exc_info()
        if notify:
            if (_dt.now()-start_time).total_seconds() >= notify_time:
                job_count = len(jobs)
                subject = 'Fyrd Wait for {} Jobs Completed'.format(job_count)
                if dispo is True:
                    message = 'Jobs completed\n'
                    subject += ', Success'
                elif isinstance(dispo, tuple):
                    message = 'Jobs {}\nTraceback:\n{}'.format(
                        repr(dispo[1]), _tb.format_tb(dispo[2])[0]
                    )
                    subject += ', {}'.format(repr(dispo[1]))
                else:
                    message = 'Some or all jobs failed\n'
                    if msg:
                        message += 'Error message:\n{}\n\n'.format(msg)
                    subject += ', Some Failed'
                message += 'Total jobs waited for: {}\n'.format(job_count)
                _notify.notify(message, notify_addr, subject)
        if isinstance(dispo, tuple):
            _reraise(*dispo)
        return dispo

    def get(self, jobs):
        """Get all results from a bunch of Job objects.

        Parameters
        ----------
        jobs : list
            List of fyrd.Job objects

        Returns
        -------
        job_results : dict
            `{job_id: Job}`

        Raises
        ------
        fyrd.ClusterError
            If any job fails or goes missing.
        """
        self.update()
        _logme.log('Queue waiting.', 'debug')
        singular = True if isinstance(jobs, self._Job) else False

        # Force into enumerated list to preserve order
        jobs = dict(enumerate(_run.listify(jobs)))
        done = {}

        # Check that all jobs are valid
        for job in jobs.values():
            if not isinstance(job, self._Job):
                raise _ClusterError('This only works with cluster job '
                                    'objects')

        # Loop through all jobs continuously trying to get outputs
        pbar = _run.get_pbar(jobs, name="Getting Job Results", unit='jobs')
        while jobs:
            for i in list(jobs):
                job = jobs[i]
                job.update()
                if not self.test_job_in_queue(job.id):
                    raise _ClusterError('Job {} not queued'.format(job.id))
                if job.state == 'completed':
                    done[i] = jobs.pop(i).get()
                    pbar.update()
                elif job.state in BAD_STATES:
                    pbar.close()
                    raise _ClusterError('Job {} failed, cannot get output'
                                        .format(job.id))
            # Block between attempts
            _sleep(self.sleep_len)
        pbar.write('Done\n')
        pbar.close()

        # Correct the order, make it the same as the input list
        results = []
        for i in sorted(done.keys()):
            results.append(done[i])
        return results[0] if singular else results

    def test_job_in_queue(self, job_id, array_id=None):
        """Check to make sure job is in self.

        Tries 12 times with 1 second between each. If found returns True,
        else False.

        Parameters
        ----------
        job_id : str
        array_id : str, optional

        Returns
        -------
        exists : bool
        """
        lgd = False
        not_found = 0
        job_id = str(job_id)
        if array_id is not None:
            array_id = str(array_id)
        if isinstance(job_id, _QueueJob):
            job_id = job_id.id
        while True:
            self._update()
            # Allow 12 seconds to elapse before job is found in queue,
            # if it is not in the queue by then, assume completion.
            if job_id in self.jobs:
                job = self.jobs[job_id]
                if array_id and array_id not in job.children:
                    return False
                return True
            else:
                if lgd:
                    _logme.log('Attempt #{}/12'.format(not_found),
                               'debug')
                else:
                    _logme.log('{} not in queue, waiting up to 12s '
                               .format(job_id) +
                               'for it to appear', 'info')
                    lgd = True
                _sleep(self.sleep_len)
                not_found += 1
                if not_found == 12:
                    _logme.log(
                        '{} not in queue, tried 12 times over 12s'
                        .format(job_id) + '. Job likely completed, ' +
                        'assuming completion, stats will be ' +
                        'unavailable.','warn'
                    )
                    return False
                continue

    def wait_to_submit(self, max_jobs=None):
        """Block until fewer running/queued jobs in queue than max_jobs.

        Parameters
        ----------
        max_jobs : int
            Override self.max_jobs for wait
        """
        count   = 50
        written = False
        while True:
            if self._can_submit(max_jobs):
                return
            if not written:
                _logme.log(('The queue is full, there are {} jobs running and '
                            '{} jobs queued. Will wait to submit, retrying '
                            'every {} seconds.')
                           .format(len(self.running), len(self.queued),
                                   self.sleep_len),
                           'info')
                written = True
            if count == 0:
                _logme.log('Still waiting to submit.', 'info')
                count = 50
            count -= 1
            _sleep(self.sleep_len)

    def update(self):
        """Refresh the list of jobs from the server, limit queries."""
        if int(_time()) - self.last_update > self.queue_update_time:
            self._update()
        else:
            _logme.log('Skipping update as last update too recent', 'debug')
        return self

    def get_jobs(self, key):
        """Return a dict of jobs where state matches key."""
        retjobs = {}
        keys = [k.lower() for k in _run.listify(key)]
        for jobid, job in self.jobs.items():
            if job.get_state() in keys:
                retjobs[jobid] = job
        return retjobs

    def get_user_jobs(self, users):
        """Filter jobs by user.

        Parameters
        ----------
        users : list
            A list of users/owners

        Returns
        -------
        dict
            A filtered job dictionary of `{job_id: QueueJob}` for all jobs
            owned by the queried users.
        """
        users = _run.listify(users)
        return {k: v for k, v in self.jobs.items() if v.owner in users}

    @property
    def users(self):
        """Return a set of users with jobs running."""
        return set([job.owner for job in self.jobs.values()])

    @property
    def job_states(self):
        """Return a list of job states for all jobs in the queue."""
        return [job.state for job in self.jobs]

    @property
    def finished(self):
        """Return a list of jobs that are neither queued nor running."""
        return {i: j for i, j in self.jobs.items() \
                if j.state not in ACTIVE_STATES}

    @property
    def bad(self):
        """Return a list of jobs that have bad or uncertain states."""
        return {i: j for i, j in self.jobs.items() \
                if j.state in BAD_STATES or j.state in UNCERTAIN_STATES}

    @property
    def active_job_count(self):
        """Return a count of all queued or running jobs, inc. array jobs."""
        self.update()
        jobcount = 0
        for j in self.get_jobs(ACTIVE_STATES):
            jobcount += j.jobcount()
        return int(jobcount)

    @property
    def can_submit(self):
        """Return True if R/Q jobs are less than max_jobs."""
        return self._can_submit()

    def _can_submit(self, max_jobs=None):
        """"Return True if R/Q jobs are less than max_jobs."""
        self.update()
        # Get max jobs
        max_jobs = int(max_jobs) if max_jobs else self.max_jobs
        if max_jobs < 4:
            max_jobs = 4
        # Fix self.max_jobs also
        if self.max_jobs < 4:
            self.max_jobs = 4
        # Update running and queued jobs can't use self.active_job_count as
        # rapid update breaks the property count and caused jobcount to be None
        jobcount = 0
        for j in self.get_jobs(ACTIVE_STATES).values():
            jobcount += j.jobcount()
        return jobcount < max_jobs

    ######################
    # Internal Functions #
    ######################

    def _update(self):
        """Refresh the list of jobs from the server.

        This is the core queue interaction function of this class.
        """
        if self._updating:
            return
        _logme.log('Queue updating', 'debug')
        # Set the update time I don't care about microseconds
        self.last_update = int(_time())

        jobs = []  # list of jobs created this session
        for [job_id, array_id, job_name, job_user, job_partition,
             job_state, job_nodelist, job_nodecount,
             job_cpus, job_exitcode] in self.batch_system.queue_parser(
                 self.user, self.partition):
            job_id = str(job_id)
            job_state = job_state.lower()
            if job_nodecount and job_cpus:
                job_threads = int(job_nodecount) * int(job_cpus)
            else:
                job_threads = None
            if job_state == 'completed' or job_state == 'failed':
                job_exitcode = job_exitcode
            else:
                job_exitcode = None

            # Get/Create the QueueJob object
            if job_id not in self.jobs:
                job = QueueJob()
            else:
                job = self.jobs[job_id]

            job.id    = job_id
            job.name  = job_name
            job.owner = job_user
            job.queue = job_partition
            job.state = job_state.lower()

            if array_id is not None:
                job.array_job = True
                cjob = QueueChild(job)
                cjob.id      = array_id
                cjob.name    = job_name
                cjob.owner   = job_user
                cjob.queue   = job_partition
                cjob.state   = job_state.lower()
                cjob.nodes   = job_nodelist
                cjob.threads = job_threads
                cjob.exitcode = job_exitcode
                job.children[array_id] = cjob
                job.state    = job.get_state()
                job.nodes    = job.get_nodelist()
                job.threads  = job.get_threads()
                job.exitcode  = job.get_exitcode()
            else:
                job.nodes   = job_nodelist
                job.threads = job_threads
                job.exitcode = job_exitcode

            # Assign the job to self.
            self.jobs[job_id] = job
            jobs.append(str(job_id))

        # We assume that if a job just disappeared it completed
        if self.jobs:
            for qjob in self.jobs.values():
                if str(qjob.id) not in jobs:
                    qjob.state = 'completed'
                    qjob.disappeared = True

    def __getattr__(self, key):
        """Make running and queued attributes dynamic."""
        key = self.batch_system.normalize_state(key.lower())
        if key == 'complete':
            key = 'completed'
        elif key == 'queued':
            key = 'pending'
        if key in ALL_STATES:
            return self.get_jobs(key)

    def __getitem__(self, key):
        """Allow direct accessing of jobs by job id."""
        if isinstance(key, self._Job):
            key = key.id
        key = str(key)
        try:
            return self.jobs[key]
        except KeyError:
            return None

    def __iter__(self):
        """Allow us to be iterable."""
        for jb in self.jobs.values():
            yield jb

    def __len__(self):
        """Length is the total job count."""
        return len(self.jobs)

    def __repr__(self):
        """For debugging."""
        self._updating = True
        if self.user:
            outstr = 'Queue<jobs:{};completed:{};pending:{};user={}>'.format(
                len(self), len(self.complete), len(self.queued), self.user)
        else:
            outstr = 'Queue<jobs:{};completed:{};pending:{};user=ALL>'.format(
                len(self), len(self.complete), len(self.queued))
        self._updating = False
        return outstr

    def __str__(self):
        """A list of keys."""
        return str(list(self.jobs.keys()))


##############################################
#  A simple class to hold jobs in the queue  #
##############################################


class _QueueJob(object):

    """A very simple class to store info about jobs in the queue.

    Attributes
    ----------
    id : int
        Job ID
    name : str
        Job name
    children : dict
        If array job, list of child job numbers
    owner : str
        User who owns the job
    threads : int
        Number of cores used by the job
    queue : str
        The queue/partition the job is running in
    state : str
        Current state of the job, normalized to slurm states
    nodes : list
        List of nodes job is running on
    exitcode : int
        Exit code of completed job
    disappeared : bool
        Job cannot be found in the queue anymore
    """

    id          = None
    name        = None
    owner       = None
    threads     = None
    queue       = None
    state       = None
    nodes       = None
    exitcode    = None
    disappeared = False
    array_job   = False
    children    = {}
    parent      = None
    _child_job  = False
    _cname      = None

    def get_state(self):
        """return the current state of the job."""
        if self.array_job:
            pending_jobs = False
            running_jobs = False
            failed_jobs  = False
            for job_info in self.children.values():
                if job_info.state == 'pending':
                    pending_jobs = True
                elif job_info.state in ACTIVE_STATES:
                    running_jobs = True
                elif job_info.state in BAD_STATES:
                    failed_jobs = True
            if running_jobs:
                return 'running'
            if pending_jobs:
                return 'pending'
            if failed_jobs:
                return 'failed'
            return 'completed'
        return self.state

    def get_nodelist(self):
        """return the current state of the job."""
        if self.array_job:
            nodelist = []
            for job_info in self.children.values():
                if job_info.nodes:
                    nodelist = nodelist + job_info.nodes
            return nodelist if nodelist else None
        return self.nodes

    def get_threads(self, state=None):
        """Return a count of how many running jobs we have."""
        states = [i.lower() for i in _run.listify(state)]
        if self.array_job:
            if state:
                return sum([j.threads for j in self.children.values()
                            if j.state in states])
            return len(self.children)
        if state:
            return self.threads if self.state in states else 0
        return self.threads

    def get_exitcode(self):
        """Return sum of exitcodes for all completed jobs."""
        if self.array_job:
            code = 0
            some_done = False
            for child in self.children.values():
                if child.state in DONE_STATES:
                    some_done = True
                    if child.exitcode:
                        code += child.exitcode
            if some_done:
                return code
            return None
        return self.exitcode

    def jobcount(self, state=None):
        """Return a count of how many running jobs we have."""
        states = [i.lower() for i in _run.listify(state)]
        if self.array_job:
            if state:
                return len([j for j in self.children.values()
                            if j.state in states])
            return len(self.children)
        if state:
            return 1 if self.state in states else 0
        return 1

    def __repr__(self):
        """Show all info."""
        if not self._child_job:
            if self.array_job:
                children = len(self.children) if self.children else None
                child_str = ':children:{0}'.format(children)
            else:
                child_str = ''
        else:
            child_str = ':parent:{0}'.format(self.parent.id)
        outstr = ("{cname}<{id}:{state}{child}" +
                  "({name},owner:{owner}," +
                  "queue:{queue},nodes:{nodes},threads:{threads}," +
                  "exitcode:{code})").format(
                      id=self.id, name=self.name, owner=self.owner,
                      queue=self.queue, nodes=self.nodes,
                      code=self.exitcode, threads=self.threads,
                      state=self.state, child=child_str, cname=self._cname
                  )
        if self.disappeared:
            outstr += 'DISAPPEARED>'
        else:
            outstr += '>'
        return outstr

    def __str__(self):
        """Print job ID."""
        return str(self.id)


class QueueJob(_QueueJob):

    """A very simple class to store info about jobs in the queue.

    Only used for torque and slurm queues.

    Attributes
    ----------
    id : int
        Job ID
    name : str
        Job name
    owner : str
        User who owns the job
    threads : int
        Number of cores used by the job
    queue : str
        The queue/partition the job is running in
    state : str
        Current state of the job, normalized to slurm states
    nodes : list
        List of nodes job is running on
    exitcode : int
        Exit code of completed job
    disappeared : bool
        Job cannot be found in the queue anymore
    array_job : bool
        This job is an array job and has children
    children : dict
        If array job, list of child job numbers
    """

    def __init__(self):
        """Initialize."""
        self._cname   = 'QueueJob'
        self.children = {}

    def __getitem__(self, key):
        """Allow direct accessing of child jobs by job id."""
        key = str(key)
        if not self.array_job:
            _logme.log('Not an array job', 'error')
            return
        return self.children[key]


class QueueChild(_QueueJob):

    """A very simple class to store info about child jobs in the queue.

    Only used for torque and slurm queues.

    Attributes
    ----------
    id : int
        Job ID
    name : str
        Job name
    owner : str
        User who owns the job
    threads : int
        Number of cores used by the job
    queue : str
        The queue/partition the job is running in
    state : str
        Current state of the job, normalized to slurm states
    nodes : list
        List of nodes job is running on
    exitcode : int
        Exit code of completed job
    disappeared : bool
        Job cannot be found in the queue anymore
    parent : QueueJob
        Backref to parent job
    """

    def __init__(self, parent):
        """Initialize with a parent."""
        self._cname   = 'QueueChild'
        self.parent = parent


################
#  Exceptions  #
################

class QueueError(Exception):

    """Simple Exception wrapper."""

    pass



# -*- coding: utf-8 -*-
"""
Modular batch system handling.

All batch system specific functions are contained within files in the
batch_systems folder.  The files must have the same name as the batch system,
and possible batch systems are set in the DEFINED_SYSTEMS set. Most batch
system functions are set in the modules in this package, but system detection
methods are hardcoded into get_cluster_environment() also.

To add new systems, create a new batch system with identical function and
classes names and return/yield values to those in an existing definition. You
will also need to update the options.py script to include keywords for your
system and the get_cluster_environment() function to include autodetection.
"""
from importlib import import_module as _import

from .. import run as _run
from .. import logme as _logme
from .. import ClusterError as _ClusterError

DEFINED_SYSTEMS = {'torque', 'slurm', 'local'}

MODE = None

# Define job states all batch systems must return one of these states
GOOD_STATES      = ['complete', 'completed', 'special_exit']
ACTIVE_STATES    = ['configuring', 'completing', 'pending',
                    'held', 'running']
BAD_STATES       = ['boot_fail', 'cancelled', 'failed', 'killed',
                    'node_fail', 'timeout', 'disappeared']
UNCERTAIN_STATES = ['preempted', 'stopped',
                    'suspended']
ALL_STATES = GOOD_STATES + ACTIVE_STATES + BAD_STATES + UNCERTAIN_STATES
DONE_STATES = GOOD_STATES + BAD_STATES


def get_batch_system(qtype=None):
    """Return a batch_system module."""
    qtype = qtype if qtype else get_cluster_environment()
    if qtype not in DEFINED_SYSTEMS:
        raise _ClusterError(
            'qtype value {0} is not recognized, '.format(qtype) +
            'should be one of {0}'.format(DEFINED_SYSTEMS)
        )
    return _import('fyrd.batch_systems.{}'.format(qtype))


#################################
#  Set the global cluster type  #
#################################


def get_cluster_environment():
    """Detect the local cluster environment and set MODE globally.

    Detect the current batch system by looking for command line utilities.
    Order is important here, so we hard code the batch system lookups.

    Paths to files can also be set in the config file.

    Returns
    -------
    MODE : str
    """
    global MODE
    from .. import conf as _conf
    conf_queue = _conf.get_option('queue', 'queue_type', 'auto')
    if conf_queue not in list(DEFINED_SYSTEMS) + ['auto']:
        _logme.log('queue_type in the config file is {}, '.format(conf_queue) +
                   'but it should be one of {}'.format(DEFINED_SYSTEMS) +
                   ' or auto. Resetting it to auto', 'warn')
        _conf.set_option('queue', 'queue_type', 'auto')
        conf_queue = 'auto'
    if conf_queue == 'auto':
        # Hardcode queue lookups here
        sbatch_cmnd = _conf.get_option('queue', 'sbatch')
        qsub_cmnd   = _conf.get_option('queue', 'qsub')
        sbatch_cmnd = sbatch_cmnd if sbatch_cmnd else 'sbatch'
        qsub_cmnd   = qsub_cmnd if qsub_cmnd else 'qsub'
        if _run.which(sbatch_cmnd):
            MODE = 'slurm'
        elif _run.which(qsub_cmnd):
            MODE = 'torque'
        else:
            from fyrd.batch_systems import local
            if local.queue_test(warn=False):
                MODE = 'local'
            else:
                MODE = None
    else:
        MODE = conf_queue
    if MODE is None:
        _logme.log('No functional batch system detected, will not be able to'
                   'run', 'error')
    elif MODE == 'local':
        _logme.log('No cluster environment detected, using multiprocessing',
                   'debug')
    else:
        _logme.log('{0} detected, using for cluster submissions'.format(MODE),
                   'debug')
    return MODE


##############################
#  Check if queue is usable  #
##############################


def check_queue(qtype=None):
    """Check if *both* MODE and qtype are valid.

    First checks the MODE global and autodetects its value, if that fails, no
    other tests are done, the qtype argument is ignored.

    After MODE is found to be a reasonable value, the queried queue is tested
    for functionality. If qtype is defined, this queue is tested, else the
    queue in MODE is tested.

    Tests are defined per batch system.

    Parameters
    ----------
    qtype : str

    Returns
    ------
    batch_system_functional : bool

    Raises
    ------
    ClusterError
        If MODE or qtype is not in DEFINED_SYSTEMS

    See Also
    --------
    get_cluster_environment : Auto detect the batch environment
    get_batch_system : Return the batch system module
    """
    if 'MODE' not in globals():
        global MODE
        MODE = get_cluster_environment()
    if not MODE:
        MODE = get_cluster_environment()
    if not MODE:
        _logme.log('Queue system not detected', 'error')
        return False
    if MODE not in DEFINED_SYSTEMS:
        raise _ClusterError(
            'MODE value {0} is not recognized, '.format(MODE) +
            'should be one of {0}'.format(DEFINED_SYSTEMS)
        )
    if qtype and qtype not in DEFINED_SYSTEMS:
        raise _ClusterError(
            'qtype value {0} is not recognized, '.format(qtype) +
            'should be one of {0}'.format(DEFINED_SYSTEMS)
        )
    qtype = qtype if qtype else MODE
    batch_system = get_batch_system(qtype)
    return batch_system.queue_test(warn=True)


# Make options easily available everywhere
from . import options



# -*- coding: utf-8 -*-
"""
Class and methods to handle Job submission.

This module only defines a single object: the Job class.
"""
import os  as _os
import sys as _sys
from uuid import uuid4 as _uuid
from time import sleep as _sleep
from datetime import datetime as _dt
from traceback import print_tb as _tb

# Try to use dill, revert to pickle if not found
import dill as _pickle
from six import reraise as _reraise
from six import text_type as _txt
from six import string_types as _str
from six import integer_types as _int

###############################################################################
#                                Our functions                                #
###############################################################################

from . import run     as _run
from . import conf    as _conf
from . import queue   as _queue
from . import logme   as _logme
from . import script_runners as _scrpts
from . import batch_systems  as _batch
from . import ClusterError   as _ClusterError
from .submission_scripts import Function as _Function
_options = _batch.options

__all__ = ['Job']

###############################################################################
#                                The Job Class                                #
###############################################################################


class Job(object):

    """Information about a single job on the cluster.

    Holds information about submit time, number of cores, the job script,
    and more.

    Below are the core attributes and methods required to use this class,
    note that this is an incomplete list.

    Attributes
    ----------
    id : str
        The ID number for the job, only set once the job has been submitted
    name : str
        The name of the job
    command : str or callable
        The function or shell script that will be submitted
    args : list
        A list of arguments to the shell script or function in command
    kwargs : dict
        A dictionary of keyword arguments to the function (not shell script) in
        command
    state : str
        A slurm-style one word description of the state of the job, one of:
            - Not_Submitted
            - queued
            - running
            - completed
            - failed
    submitted : bool
    written : bool
    done : bool
    running : bool
    dependencies : list
        A list of dependencies associated with this job
    out : str
        The output of the function or a copy of stdout for a script
    stdout : str
        Any output to STDOUT
    stderr : str
        Any output to STDERR
    exitcode : int
        The exitcode of the running processes (the script runner if the Job is
        a function).
    submit_time : datetime
        A datetime object for the time of submission
    start : datetime
        A datetime object for time execution started on the remote node.
    end : datetime
        A datetime object for time execution ended on the remote node.
    runtime : timedelta
        A timedelta object containing runtime.
    files : list
        A list of script files associated with this job
    nodes : list
        A list of nodes associated with this job
    modules : list
        A list of modules associated with this job
    clean_files : bool
        If True, auto-delete script and function files on job completion
    clean_outputs : bool
        If True, auto-delete script outputs and error files on job completion
    kwds : dict
        Keyword arguments to the batch system (e.g. mem, cores, walltime), this
        is initialized by taking every additional keyword argument to the Job.
        e.g. Job('echo hi', profile=large, walltime='00:20:00', mem='2GB') will
        result in kwds containing {walltime: '00:20:00', mem: '2GB'}. There is
        **no need to alter this manually**.
    submit_args : list
        List of parsed submit arguments that will be passed at runtime to the
        submit function. **Generated within the Job object**, no need to set
        manually, use the `kwds` attribute instead.

    Methods
    -------
    initialize()
        Use attributes to prep job for running
    gen_scripts()
        Create script files (but do not write them)
    write(overwrite=True)
        Write scripts to files
    submit(wait_on_max_queue=True)
        Submit the job if it is ready and the queue is sufficiently open.
    resubmit(wait_on_max_queue=True)
        Clean all internal states with `scrub()` and then resubmit
    kill(confirm=True)
        Immediately kill the currently running job
    clean(delete_outputs=True, get_outputs=True)
        Delete any files created by this object
    scrub(confirm=True)
        Clean everything and reset to an unrun state.
    update(fetch_info=True)
        Update our status from the queue
    wait()
        Block until the job is done
    get()
        Block until the job is done and then return the output (stdout if job
        is a script), by default saves all outputs to self (i.e. .out, .stdout,
        .stderr) and deletes all intermediate files before returning. If `save`
        argument is `False`, does not delete the output files by default.

    Notes
    -----
    Printing or reproducing the class will display detailed job information.

    Both `wait()` and `get()` will update the queue every few seconds
    (defined by the queue_update item in the config) and add queue information
    to the job as they go.

    If the job disappears from the queue with no information, it will be listed
    as 'completed'.

    All jobs have a .submission attribute, which is a Script object containing
    the submission script for the job and the file name, plus a 'written' bool
    that checks if the file exists.

    In addition, some batch systems (e.g. SLURM) have an .exec_script
    attribute, which is a Script object containing the shell command to run.
    This difference is due to the fact that some SLURM systems execute multiple
    lines of the submission file at the same time.

    Finally, if the job command is a function, this object will also contain a
    `.function` attribute, which contains the script to run the function.
    """

    id            = None
    name          = None
    suffix        = None
    submitted     = False
    written       = False
    found         = False
    disappeared   = False
    submit_time   = None
    state         = None
    kind          = None

    # Arguments
    kwds          = None
    kwargs        = None
    submit_args   = None

    # Runtime
    nodes         = None
    cores         = None
    modules       = None

    # Files
    outfile       = None
    errfile       = None

    # Scripts
    submission    = None
    exec_script   = None
    function      = None
    imports       = None

    # Dependencies
    dependencies  = None

    # Pickled output file for functions
    poutfile      = None

    # Holds queue information in torque and slurm
    queue_info    = None

    # Output tracking
    _got_out      = False
    _got_stdout   = False
    _got_stderr   = False
    _got_exitcode = False
    _found_files  = False
    _out          = None
    _stdout       = None
    _stderr       = None
    _exitcode     = None

    # Time tracking
    _got_times    = False
    start         = None
    end           = None

    # Track update status
    _updating     = False

    # Track preparations
    initialized   = False
    scripts_ready = False
    _kwargs       = None

    # Auto Cleaning
    clean_files   = _conf.get_option('jobs', 'clean_files')
    clean_outputs = _conf.get_option('jobs', 'clean_outputs')

    def __init__(self, command, args=None, kwargs=None, name=None, qtype=None,
                 profile=None, **kwds):
        """Initialization function arguments.

        Parameters
        ----------
        command : function/str
            The command or function to execute.
        args : tuple/dict, optional
            Optional arguments to add to command, particularly useful for
            functions.
        kwargs : dict, optional
            Optional keyword arguments to pass to the command, only used for
            functions.
        name : str, optional
            Optional name of the job. If not defined, guessed. If a job of the
            same name is already queued, an integer job number (not the queue
            number) will be added, ie.  <name>.1
        qtype : str, optional
            Override the default queue type
        profile : str, optional
            The name of a profile saved in the conf
        kwds
            *All other keywords are parsed into cluster keywords by the options
            system.* For available keywords see `fyrd.option_help()`
        """

        ########################
        #  Sanitize arguments  #
        ########################
        _logme.log('Args pre-check: {}'.format(kwds), 'debug')
        kwds = _options.check_arguments(kwds)
        _logme.log('Args post-check: {}'.format(kwds), 'debug')

        # Create a unique short UUID for this job
        self.uuid = str(_uuid()).split('-')[0]

        # Path handling
        [kwds, self.runpath,
         self.outpath, self.scriptpath] = _conf.get_job_paths(kwds)

        # Save command
        self.command = command
        self.args    = args
        self.kwargs  = kwargs
        self.profile = profile

        # Get environment
        if not _batch.MODE:
            _batch.MODE = _batch.get_cluster_environment()
        self.qtype = qtype if qtype else _batch.MODE
        self.queue = _queue.Queue(user='self', qtype=self.qtype)
        self.batch = _batch.get_batch_system(self.qtype)
        self.state = 'Not_Submitted'

        # Save keywords for posterity and parsing
        self.kwds = kwds

        self.name = self._update_name(name)

        self.initialize()

    ##########################################################################
    #                             Public Methods                             #
    ##########################################################################

    ################
    #  Properties  #
    ################

    @property
    def files(self):
        """Build a list of files associated with this class."""
        files = [self.submission]
        if self.kind == 'script':
            files.append(self.exec_script)
        if self.kind == 'function':
            files.append(self.function)
        return files

    @property
    def runtime(self):
        """Return the runtime."""
        if not self.done:
            _logme.log('Cannot get runtime as not yet complete.' 'warn')
            return None
        if not self.start:
            self.get_times()
        return self.end-self.start

    @property
    def done(self):
        """Check if completed or not.

        Updates the Job and Queue.

        Returns
        -------
        done : bool
        """
        # We have the same statement twice to try and avoid updating.
        if self.state in _batch.DONE_STATES:
            return True
        if not self._updating:
            self.update()
            if self.state in _batch.DONE_STATES:
                return True
        return False

    @property
    def running(self):
        """Check if running or not.

        Updates the Job and Queue.

        Returns
        -------
        running : bool
        """
        # We have the same statement twice to try to avoid updating.
        if self.state in _batch.ACTIVE_STATES:
            return True
        if not self._updating:
            self.update()
            if self.state in _batch.ACTIVE_STATES:
                return True
        return False

    @property
    def outfiles(self):
        """A list of all outfiles associated with this Job."""
        outfiles = [self.outfile, self.errfile]
        if self.poutfile:
            outfiles.append(self.poutfile)
        return outfiles

    @property
    def incomplete_outfiles(self):
        """A list of all outfiles that haven't already been fetched."""
        outfiles = []
        if self.outfile and not self._got_stdout:
            outfiles.append(self.outfile)
        if self.errfile and not self._got_stderr:
            outfiles.append(self.errfile)
        if self.poutfile and not self._got_out:
            outfiles.append(self.poutfile)
        return outfiles

    @property
    def exitcode(self):
        """Return exitcode."""
        return self.get_exitcode()

    @property
    def code(self):
        """Return exitcode."""
        return self.get_exitcode()

    @property
    def out(self):
        """Return output."""
        return self.get_output()

    @property
    def stdout(self):
        """Return output."""
        return self.get_stdout()

    @property
    def stderr(self):
        """Return stderr."""
        return self.get_stderr()

    @property
    def err(self):
        """Return stderr."""
        return self.get_stderr()

    ###############################
    #  Core Job Handling Methods  #
    ###############################

    def initialize(self):
        """Make self runnable using set attributes."""
        kwds = self.kwds

        # Override autoclean state (set in config file)
        if 'clean_files' in kwds:
            self.clean_files = kwds.pop('clean_files')
        if 'clean_outputs' in kwds:
            self.clean_outputs = kwds.pop('clean_outputs')

        # Set suffix
        self.suffix = kwds.pop('suffix') if 'suffix' in kwds \
                      else _conf.get_option('jobs', 'suffix')

        # Merge in profile, this includes all args from the DEFAULT profile
        # as well, ensuring that those are always set at a minumum.
        profile = self.profile if self.profile else 'DEFAULT'
        prof = _conf.get_profile(profile)
        if not prof:
            raise _ClusterError('No profile found for {}'.format(profile))
        for k,v in prof.args.items():
            if k not in kwds:
                kwds[k] = v

        # Use the default profile as a backup if any arguments missing
        default_args = _conf.DEFAULT_PROFILES['DEFAULT']
        default_args.update(_conf.get_profile('DEFAULT').args)
        for opt, arg in default_args.items():
            if opt not in kwds:
                _logme.log('{} not in kwds, adding from default: {}:{}'
                           .format(opt, opt, arg), 'debug')
                kwds[opt] = arg

        # Set modules
        self.modules = kwds.pop('modules') if 'modules' in kwds else None
        if self.modules:
            self.modules = _run.opt_split(self.modules, (',', ';'))

        # Make sure args are a tuple
        if self.args:
            self.args = tuple(_run.listify(self.args))

        # In case cores are passed as None
        if 'nodes' not in kwds:
            kwds['nodes'] = default_args['nodes']
        if 'cores' not in kwds:
            kwds['cores'] = default_args['cores']
        self.nodes = kwds['nodes']
        self.cores = kwds['cores']

        # Set output files
        if 'outfile' in kwds:
            pth, fle = _os.path.split(kwds['outfile'])
            if not pth:
                pth = self.outpath
            kwds['outfile'] = _os.path.join(pth, fle)
        else:
            kwds['outfile'] = _os.path.join(
                self.outpath, '.'.join([self.name, self.suffix, 'out']))
        if 'errfile' in kwds:
            pth, fle = _os.path.split(kwds['errfile'])
            if not pth:
                pth = self.outpath
            kwds['errfile'] = _os.path.join(pth, fle)
        else:
            kwds['errfile'] = _os.path.join(
                self.outpath, '.'.join([self.name, self.suffix, 'err']))
        self.outfile = kwds['outfile']
        self.errfile = kwds['errfile']

        # Check and set dependencies
        if 'depends' in kwds:
            dependencies = _run.listify(kwds.pop('depends'))
            self.dependencies = []
            errmsg = 'Dependencies must be number, numeric string or Job'
            for dependency in dependencies:
                if not isinstance(dependency, (_str, _txt, Job)):
                    raise _ClusterError(errmsg)
                self.dependencies.append(dependency)

        # Save parsed keywords as _kwargs
        self._kwargs = kwds

        self.initialized = True

        return self

    def gen_scripts(self):
        """Create the script objects from the set parameters."""
        if not self.initialized:
            self.initialize()

        ######################################
        #  Command and Function Preparation  #
        ######################################
        command = self.command
        args    = self.args
        kwargs  = self.kwargs  # Not self._kwargs
        name    = self._update_name()
        kwds    = self._kwargs

        # Get imports
        imports = kwds.pop('imports') if 'imports' in kwds else None

        # Get syspaths
        syspaths = kwds.pop('syspaths') if 'syspaths' in kwds else None

        # Split out sys.paths from imports and set imports in self
        if imports:
            self.imports = []
            syspaths = syspaths if syspaths else []
            for i in imports:
                if i.startswith('sys.path.append')\
                        or i.startswith('sys.path.insert'):
                    syspaths.append(i)
                else:
                    self.imports.append(i)

        # Function specific initialization
        if callable(command):
            self.kind = 'function'
            script_file = _os.path.join(
                self.scriptpath, '{}_func.{}.py'.format(name, self.suffix)
                )
            self.poutfile = self.outfile + '.func.pickle'
            self.function = _Function(
                file_name=script_file, function=command, args=args,
                kwargs=kwargs, imports=self.imports, syspaths=syspaths,
                outfile=self.poutfile
            )
            # Collapse the _command into a python call to the function script
            executable = '#!/usr/bin/env python{}'.format(
                _sys.version_info.major) if _conf.get_option(
                    'jobs', 'generic_python') else _sys.executable

            command = '{} {}'.format(executable, self.function.file_name)
            args = None
        else:
            self.kind = 'script'
            self.poutfile = None

        # Collapse args into command
        command = command + ' '.join(args) if args else command

        #####################
        #  Script Creation  #
        #####################

        # Build execution wrapper with modules
        modstr  = ''
        if self.modules:
            for module in self.modules:
                modstr += 'module load {}\n'.format(module)

        # Add all of the keyword arguments at once
        opt_string, submit_args = _options.options_to_string(kwds, self.qtype)
        precmd = opt_string + '\n\n' + modstr
        self.submit_args = submit_args

        # Create queue-dependent scripts
        self.submission, self.exec_script = self.batch.gen_scripts(
            self, command, args, precmd, modstr
        )

        self.scripts_ready = True

        return self

    def write(self, overwrite=True):
        """Write all scripts.

        Parameters
        ----------
        overwrite : bool, optional
            Overwrite existing files, defaults to True.

        Returns
        -------
        self : Job
        """
        if not self.scripts_ready:
            self.gen_scripts()
        _logme.log('Writing files, overwrite={}'.format(overwrite), 'debug')
        self.submission.write(overwrite)
        if self.exec_script:
            self.exec_script.write(overwrite)
        if self.function:
            self.function.write(overwrite)
        self.written = True
        return self

    def submit(self, wait_on_max_queue=True, additional_keywords=None,
               max_jobs=None):
        """Submit this job.

        To disable max_queue_len, set it to 0. None will allow override by
        the default settings in the config file, and any positive integer will
        be interpretted to be the maximum queue length.

        Parameters
        ----------
        wait_on_max_queue : bool, optional
            Block until queue limit is below the maximum before submitting.
        additional_keywords : dict, optional
            Pass this dictionary to the batch system submission function,
            not necessary.
        max_jobs : int, optional
            Override the maximum number of jobs to wait for

        Returns
        -------
        self : Job
        """
        if self.submitted:
            _logme.log('Not submitting, already submitted.', 'warn')
            return self

        if not self.written:
            self.write()

        # Check dependencies
        dependencies = []
        if self.dependencies:
            for depend in self.dependencies:
                if isinstance(depend, Job):
                    if not depend.id:
                        _logme.log(
                            'Cannot submit job as dependency {} '
                            .format(depend) + 'has not been submitted',
                            'error'
                        )
                        return self
                    dependencies.append(str(depend.id))
                else:
                    dependencies.append(str(depend))

        # Wait on the queue if necessary
        if wait_on_max_queue:
            if not self._updating:
                self.update()
            self.queue.wait_to_submit(max_jobs)

        # Only include queued or running dependencies
        self.queue._update()  # Force update
        depends = []
        for depend in dependencies:
            dep_check = self.queue.check_dependencies(depend)
            if dep_check == 'absent':
                _logme.log(
                    'Cannot submit job as dependency {} '
                    .format(depend) + 'is not in the queue',
                    'error'
                )
                return self
            elif dep_check == 'good':
                _logme.log(
                    'Dependency {} is complete, skipping'
                    .format(depend), 'debug'
                )
            elif dep_check == 'bad':
                _logme.log(
                    'Cannot submit job as dependency {} '
                    .format(depend) + 'has failed',
                    'error'
                )
                return self
            elif dep_check == 'active':
                if self.queue.jobs[depend].state == 'completeing':
                    continue
                _logme.log('Dependency {} is {}, adding to deps'
                           .format(depend, self.queue.jobs[depend].state),
                           'debug')
                depends.append(depend)
            else:
                # This shouldn't happen ever
                raise _ClusterError('fyrd.queue.Queue.check_dependencies() ' +
                                    'returned an unrecognized value {0}'
                                    .format(dep_check))

        self.id = self.batch.submit(
            self.submission.file_name,
            dependencies=depends,
            job=self, args=self.submit_args,
            kwds=additional_keywords
        )

        self.submitted = True
        self.submit_time = _dt.now()
        self.state = 'submitted'

        if not self.submitted:
            raise _ClusterError('Submission appears to have failed, this '
                                "shouldn't happen")

        return self

    def resubmit(self, wait_on_max_queue=True, cancel_running=None):
        """Attempt to auto resubmit, deletes prior files.

        Parameters
        ----------
        wait_on_max_queue : bool, optional
            Block until queue limit is below the maximum before submitting.
        cancel_running : bool or None, optional
            If the job is currently running, cancel it before resubmitting.
            If None (default), will ask the user.

        To disable max_queue_len, set it to 0. None will allow override by
        the default settings in the config file, and any positive integer will
        be interpretted to be the maximum queue length.

        Returns
        -------
        self : Job
        """
        if self.running:
            if cancel_running is None:
                cancel_running = _run.get_yesno(
                    'Job currently running, cancel before resubmitting?', 'y'
                )
            if cancel_running:
                self.kill(confirm=False)
        self.scrub(confirm=False)
        # Rerun
        self.initialize()
        self.gen_scripts()
        self.write()
        return self.submit(wait_on_max_queue)

    def kill(self, confirm=True):
        """Kill the running job.

        Parameters
        ----------
        confirm : bool, optional

        Returns
        -------
        self : Job
        """
        if not self.submitted:
            _logme.log('Job not submitted, cannot kill', 'warn')
            return self
        if self.done:
            _logme.log('Job completed, cannot kill', 'warn')
            return self
        if confirm:
            if not _run.get_yesno(
                'This will terminate the running job, continue?', 'n'
            ):
                return self
        self.batch.kill(self.id)
        return self

    def clean(self, delete_outputs=None, get_outputs=True):
        """Delete all scripts created by this module, if they were written.

        Parameters
        ----------
        delete_outputs : bool, optional
            also delete all output and err files, but get their contents first.
        get_outputs : bool, optional
            if delete_outputs, save outputs before deleting.

        Returns
        -------
        self : Job
        """
        _logme.log('Cleaning outputs, delete_outputs={}'
                   .format(delete_outputs), 'debug')
        if not isinstance(delete_outputs, bool):
            delete_outputs = self.clean_outputs
        assert isinstance(delete_outputs, bool)
        for jobfile in [self.submission, self.exec_script, self.function]:
            if jobfile:
                jobfile.clean()
        if delete_outputs:
            _logme.log('Deleting output files.', 'debug')
            if get_outputs:
                self.fetch_outputs(delete_files=True)
            for f in self.outfiles:
                if _os.path.isfile(f):
                    _logme.log('Deleteing {}'.format(f), 'debug')
                    _os.remove(f)
        return self

    def scrub(self, confirm=True):
        """Clean everything and reset to an unrun state.

        Parameters
        ----------
        confirm : bool, optional
            Get user input before proceeding

        Returns
        -------
        self : Job
        """
        msg = ("This will delete all outputs stored in this job, as well "
               "as all output files, job files, and scripts. Are you sure "
               "you want to do this?")
        if confirm:
            _run.get_yesno(msg, default='n')
        # Clean old set
        self.clean(delete_outputs=True)
        # Reset runtime attributes
        self.initialized   = False
        self.scripts_ready = False
        self.written       = False
        self.submitted     = False
        self.id            = None
        self.found         = False
        self.queue_info    = None
        self.state         = 'Not_Submitted'
        self._got_out      = False
        self._got_stdout   = False
        self._got_stderr   = False
        self._got_exitcode = False
        self._out          = None
        self._stdout       = None
        self._stderr       = None
        self._exitcode     = None
        self._got_times    = False
        self._updating     = False
        self._found_files  = False
        self.start         = None
        self.end           = None
        return self.update()

    ######################
    #  Queue Management  #
    ######################

    def update(self, fetch_info=True):
        """Update status from the queue.

        Parameters
        ----------
        fetch_info : bool, optional
            Fetch basic job info if complete.

        Returns
        -------
        self : Job
        """
        if not self._updating:
            self._update(fetch_info)
        else:
            _logme.log('Already updating, aborting.', 'debug')
        return self

    def update_queue_info(self):
        """Set (and return) queue_info from the queue even if done."""
        _logme.log('Updating queue_info', 'debug')
        queue_info1 = self.queue[self.id]
        self.queue.update()
        queue_info2 = self.queue[self.id]
        if queue_info2:
            self.queue_info = queue_info2
        elif queue_info1:
            self.queue_info = queue_info1
        elif self.queue_info is None and self.submitted:
            _logme.log('Cannot find self in the queue and queue_info is empty',
                       'warn')
        return self.queue_info

    #################################
    #  Output Handling and Waiting  #
    #################################

    def wait(self):
        """Block until job completes.

        Returns
        -------
        success : bool or str
            True if exitcode == 0, False if not, 'disappeared' if job lost from
            queue.
        """
        if not self.submitted:
            if _conf.get_option('jobs', 'auto_submit'):
                _logme.log('Auto-submitting as not submitted yet', 'debug')
                self.submit()
            else:
                _logme.log('Cannot wait for result as job has not been ' +
                           'submitted', 'warn')
                return False
        self.update(fetch_info=False)
        if not self.done:
            _logme.log('Waiting for self {}'.format(self.name), 'debug')
            status = self.queue.wait(self.id, return_disp=True)
            if status == 'disappeared':
                self.state = status
            elif status is not True:
                return False
            else:
                if not self._updating:
                    self.update()
                if self.get_exitcode(update=False) != 0:
                    _logme.log('Job failed with exitcode {}'
                               .format(self.exitcode), 'debug')
                    return False
        if self._wait_for_files(caution_message=False):
            if not self._updating:
                self.update()
            if self.state == 'disappeared':
                _logme.log('Job files found for disappered job, assuming '
                           'success', 'info')
                return 'disappeared'
            return True
        else:
            if self.state == 'disappeared':
                _logme.log('Disappeared job has no output files, assuming '
                           'failure', 'error')
            return False

    def get(self, save=True, cleanup=None, delete_outfiles=None,
            del_no_save=None, raise_on_error=True):
        """Block until job completed and return output of script/function.

        By default saves all outputs to this class and deletes all intermediate
        files.

        Parameters
        ----------
        save : bool, optional
            Save all outputs to the class also (advised)
        cleanup : bool, optional
            Clean all intermediate files after job completes.
        delete_outfiles : bool, optional
            Clean output files after job completes.
        del_no_save : bool, optional
            Delete output files even if `save` is `False`
        raise_on_error : bool, optional
            If the returned output is an Exception, raise it.

        Returns
        -------
        str
            Function output if Function, else STDOUT
        """
        _logme.log(('Getting outputs, cleanup={}, autoclean={}, '
                    'delete_outfiles={}').format(
                        cleanup, self.clean_files, delete_outfiles
                    ), 'debug')
        # Wait for queue
        status = self.wait()
        if status is not True:
            if status == 'disappeared':
                msg = 'Job disappeared from queue'
                _logme.log(msg + ', attempting to get '
                           'outputs', 'debug')
            else:
                msg = 'Wait failed'
                _logme.log(msg + ', attempting to get outputs anyway',
                           'debug')
            try:
                self.fetch_outputs(save=save, delete_files=False,
                                   get_stats=False)
            except IOError:
                _logme.log(msg + ' and files could not be found, job must '
                           'have failed', 'error')
                if raise_on_error:
                    raise
                return
            if status != 'disappeared':
                return
        else:
            # Get output
            _logme.log('Wait complete, fetching outputs', 'debug')
            self.fetch_outputs(save=save, delete_files=False)
        out = self.out if save else self.get_output(save=save, update=False)
        if isinstance(out, tuple) and issubclass(out[0], Exception):
            if raise_on_error:
                _reraise(*out)
            else:
                _logme.log('Job failed with exception {}'.format(out))
                print(_tb(out[2]))
                return out
        # Cleanup
        if cleanup is None:
            cleanup = self.clean_files
        else:
            assert isinstance(cleanup, bool)
        if delete_outfiles is None:
            delete_outfiles = self.clean_outputs
        if save is False:
            delete_outfiles = del_no_save if del_no_save is not None else False
        if cleanup:
            self.clean(delete_outputs=delete_outfiles)
        return out

    def get_output(self, save=True, delete_file=None, update=True,
                   raise_on_error=True):
        """Get output of function or script.

        This is the same as stdout for a script, or the function output for
        a function.

        By default, output file is kept unless delete_file is True or
        self.clean_files is True.

        Parameters
        ----------
        save : bool, optional
            Save the output to self.out, default True.  Would be a good idea to
            set to False if the output is huge.
        delete_file : bool, optional
            Delete the output file when getting
        update : bool, optional
            Update job info from queue first.
        raise_on_error : bool, optional
            If the returned output is an Exception, raise it.

        Returns
        -------
        output : anything
            The output of the script or function. Always a string if script.
        """
        _logme.log(('Getting output, save={}, clean_files={}, '
                    'delete_file={}').format(
                        save, self.clean_files, delete_file
                    ), 'debug')
        if delete_file is None:
            delete_file = self.clean_outputs
        if self.kind == 'script':
            return self.get_stdout(save=save, delete_file=delete_file,
                                   update=update)
        if self.done and self._got_out:
            _logme.log('Getting output from _out', 'debug')
            return self._out
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self._wait_for_files()
        else:
            _logme.log('Cannot get pickled output before job completes',
                       'warn')
            return None
        _logme.log('Getting output from {}'.format(self.poutfile), 'debug')
        if _os.path.isfile(self.poutfile):
            with open(self.poutfile, 'rb') as fin:
                out = _pickle.load(fin)
            if delete_file is True or self.clean_files is True:
                _logme.log('Deleting {}'.format(self.poutfile),
                           'debug')
                _os.remove(self.poutfile)
            if save:
                self._out = out
                self._got_out = True
            if _run.is_exc(out):
                _logme.log('{} failed with exception {}'.format(self, out[1]),
                           'error')
                if raise_on_error:
                    _reraise(*out)
            return out
        else:
            _logme.log('No file at {} even though job has completed!'
                       .format(self.poutfile), 'critical')
            raise IOError('File not found: {}'.format(self.poutfile))

    def get_stdout(self, save=True, delete_file=None, update=True):
        """Get stdout of function or script, same for both.

        By default, output file is kept unless delete_file is True or
        self.clean_files is True.

        Also sets self.start and self.end from the contents of STDOUT if
        possible.

        Returns
        -------
        save : bool, optional
            Save the output to self.stdout, default True.  Would be a good idea
            to set to False if the output is huge.
        delete_file : bool, optional
            Delete the stdout file when getting
        update : bool, optional
            Update job info from queue first.

        Returns
        -------
        str
            The contents of STDOUT, with runtime info and trailing newline
            removed.
        """
        if delete_file is None:
            delete_file = self.clean_outputs
        _logme.log(('Getting stdout, save={}, clean_files={}, '
                    'delete_file={}').format(
                        save, self.clean_files, delete_file
                    ), 'debug')
        if self.done and self._got_stdout:
            _logme.log('Getting stdout from _stdout', 'debug')
            return self._stdout
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self._wait_for_files()
        else:
            _logme.log('Job not done, attempting to get current STDOUT ' +
                       'anyway', 'info')
        _logme.log('Getting stdout from {}'.format(self._kwargs['outfile']),
                   'debug')
        if _os.path.isfile(self._kwargs['outfile']):
            with open(self._kwargs['outfile']) as fin:
                stdout = fin.read()
            if stdout:
                stdouts = stdout.strip().split('\n')
                if len(stdouts) < 3 or stdouts[-3] != 'Done':
                    _logme.log('STDOUT incomplete, returning as is', 'info')
                    return stdout
                if self.done:
                    self.get_times(update=False, stdout=stdout)
                    self.get_exitcode(update=False, stdout=stdout)
                stdout  = '\n'.join(stdouts[2:-3]) + '\n'
            if delete_file is True or self.clean_files is True:
                _logme.log('Deleting {}'.format(self._kwargs['outfile']),
                           'debug')
                _os.remove(self._kwargs['outfile'])
            if save:
                self._stdout = stdout
                if self.done:
                    self._got_stdout = True
            return stdout
        else:
            _logme.log('No file at {}, cannot get stdout'
                       .format(self._kwargs['outfile']), 'warn')
            return None

    def get_stderr(self, save=True, delete_file=None, update=True):
        """Get stderr of function or script, same for both.

        By default, output file is kept unless delete_file is True or
        self.clean_files is True.

        Parameters
        ----------
        save : bool, optional
            Save the output to self.stdout, default True.  Would be a good idea
            to set to False if the output is huge.
        delete_file : bool, optional
            Delete the stdout file when getting
        update : bool, optional
            Update job info from queue first.

        Returns
        -------
        str
            The contents of STDERR, with trailing newline removed.
        """
        if delete_file is None:
            delete_file = self.clean_outputs
        _logme.log(('Getting stderr, save={}, clean_files={}, '
                    'delete_file={}').format(
                        save, self.clean_files, delete_file
                    ), 'debug')
        if self.done and self._got_stderr:
            _logme.log('Getting stderr from _stderr', 'debug')
            return self._stderr
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self._wait_for_files()
        else:
            _logme.log('Job not done, attempting to get current STDERR ' +
                       'anyway', 'info')
        _logme.log('Getting stderr from {}'.format(self._kwargs['errfile']),
                   'debug')
        if _os.path.isfile(self._kwargs['errfile']):
            with open(self._kwargs['errfile']) as fin:
                stderr = fin.read()
            if delete_file is True or self.clean_files is True:
                _logme.log('Deleting {}'.format(self._kwargs['errfile']),
                           'debug')
                _os.remove(self._kwargs['errfile'])
            if save:
                self._stderr = stderr
                if self.done:
                    self._got_stderr = True
            return stderr
        else:
            _logme.log('No file at {}, cannot get stderr'
                       .format(self._kwargs['errfile']), 'warn')
            return None

    def get_times(self, update=True, stdout=None):
        """Get stdout of function or script, same for both.

        Sets self.start and self.end from the contents of STDOUT if
        possible.

        Parameters
        ----------
        update : bool, optional
            Update job info from queue first.
        stdout : str, optional
            Pass existing stdout for use

        Returns
        -------
        start : datetime.datetime
        end : datetime.datetime
        """
        _logme.log('Getting times', 'debug')
        if self.done and self._got_times:
            _logme.log('Getting times from self.start, self.end', 'debug')
            return self.start, self.end
        if update and not self._updating and not self.done:
            self.update()
        if self.done:
            if update:
                self._wait_for_files()
        else:
            _logme.log('Cannot get times until job is complete.', 'warn')
            return None, None
        _logme.log('Getting times from {}'.format(self._kwargs['outfile']),
                   'debug')
        if not stdout:
            if _os.path.isfile(self._kwargs['outfile']):
                with open(self._kwargs['outfile']) as fin:
                    stdout = fin.read()
            else:
                _logme.log('No file at {}, cannot get times'
                           .format(self._kwargs['outfile']), 'warn')
                return None
        stdouts = stdout.strip().split('\n')
        if len(stdouts) < 3 or stdouts[-3] != 'Done':
            _logme.log('STDOUT incomplete, cannot get times', 'warn')
            return None

        # Get times
        timefmt = '%y-%m-%d-%H:%M:%S'
        try:
            self.start = _dt.strptime(stdouts[0], timefmt)
            self.end   = _dt.strptime(stdouts[-1], timefmt)
        except ValueError as err:
            _logme.log('Time parsing failed with value error; ' +
                       '{}. '.format(err) + 'This may be because you ' +
                       'are using the script running that does not ' +
                       'include time tracking', 'debug')
        self._got_times = True
        return self.start, self.end

    def get_exitcode(self, update=True, stdout=None):
        """Try to get the exitcode.

        Parameters
        ----------
        update : bool, optional
            Update job info from queue first.
        stdout : str, optional
            Pass existing stdout for use

        Returns
        -------
        exitcode : int
        """
        _logme.log('Getting exitcode', 'debug')
        if self.done and self._got_exitcode:
            _logme.log('Getting exitcode from _exitcode', 'debug')
            return self._exitcode
        if update and not self._updating and not self.done:
            self.update()
        if not self.done:
            _logme.log('Job is not complete, no exit code yet', 'info')
            return None
        if self.state == 'disappeared':
            _logme.log('Cannot get exitcode for disappeared job', 'debug')
            return 0

        code = None

        if not stdout and _os.path.isfile(self._kwargs['outfile']):
            with open(self._kwargs['outfile']) as fin:
                stdout = fin.read()

        if stdout:
            stdouts = stdout.strip().split('\n')
            if len(stdouts) > 3 and stdouts[-3] == 'Done':
                if stdouts[-2].startswith('Code: '):
                    code = int(stdouts[-2].split(':')[-1].strip())

        if code is None:
            _logme.log('Getting exitcode from queue', 'debug')
            if not self.queue_info:
                self.queue_info = self.queue[self.id]
            if hasattr(self.queue_info, 'exitcode'):
                code = self.queue_info.exitcode

        if code is None:
            _logme.log('Failed to get exitcode for job', 'warn')
            return None

        self._exitcode     = code
        self._got_exitcode = True

        if code != 0:
            self.state = 'failed'
            _logme.log('Job {} failed with exitcode {}'
                       .format(self.name, code), 'error')

        return code

    def fetch_outputs(self, save=True, delete_files=None, get_stats=True):
        """Save all outputs in their current state. No return value.

        This method does not wait for job completion, but merely gets the
        outputs. To wait for job completion, use `get()` instead.

        Parameters
        ----------
        save : bool, optional
            Save all outputs to the class also (advised)
        delete_files : bool, optional
            Delete the output files when getting, only used if save is True
        get_stats : bool, optional
            Try to get exitcode.
        """
        _logme.log('Saving outputs to self, delete_files={}'
                   .format(delete_files), 'debug')
        if not self._updating:
            self.update()
        if delete_files is None:
            delete_files = self.clean_outputs
        if not self._got_exitcode and get_stats:
            self.get_exitcode(update=False)
        if not self._got_times:
            self.get_times(update=False)
        if save:
            self.get_output(save=True, delete_file=delete_files, update=False)
            self.get_stdout(save=True, delete_file=delete_files, update=False)
            self.get_stderr(save=True, delete_file=delete_files, update=False)

    ##############################
    #  Minor management methods  #
    ##############################

    def get_keywords(self):
        """Return a list of the keyword arguments used to make the job."""
        return self.kwds

    def set_keywords(self, kwds, replace=False):
        """Set the job keywords, just updates self.kwds.

        Parameters
        ----------
        kwds : dict
            Set of valid arguments.
        replace : bool, optional
            Overwrite the keword arguments instead of updating.
        """
        kwds = _options.check_arguments(kwds)
        if replace:
            self.kwds = kwds
        else:
            for key, value in kwds.items():
                self.kwds[key] = value

    ###############
    #  Internals  #
    ###############

    def _update(self, fetch_info=True):
        """Update status from the queue.

        Parameters
        ----------
        fetch_info : bool, optional
            Fetch basic job info if complete.
        """
        _logme.log('Updating job.', 'debug')
        self._updating = True
        if self.done or not self.submitted:
            self._updating = False
            return
        self.queue.update()
        if self.submitted and self.id:
            queue_info = self.queue[self.id]
            if queue_info:
                assert self.id == queue_info.id
                self.found = True
                self.queue_info = queue_info
                self.state = self.queue_info.state
            elif self.found:
                _logme.log('Job appears to have disappeared, waiting for '
                           'reappearance, this may take a while', 'warn')
                status = self.wait()
                if status == 'disappeared':
                    _logme.log('Job disappeared, but the output files are '
                               'present assuming completion', 'info')
                    self.state = 'completed'
                    self.disappeared = True
                elif not status:
                    _logme.log('Job appears to have failed and disappeared',
                               'error')
            # If job not found after 30 seconds, assume trouble, check for
            # completion
            elif self.submitted and (_dt.now()-self.submit_time).seconds > 360:
                if self._wait_for_files(btme=4, caution_message=False):
                    self.state = 'completed'
                    self.disappeared = True
                    _logme.log('Job never appeared in the queue, but '
                               'outfiles still exist, assuming completion.',
                               'warn')
                else:
                    self.state = 'failed'
                    self.disappeared = True
                    s = (_dt.now()-self.submit_time).seconds
                    _logme.log('Job not in queue after {} seconds '.format(s) +
                               'of searching and no outputs found, assuming '
                               'failure.', 'error')
            elif self.submitted and (_dt.now()-self.submit_time).seconds > 30:
                if self._wait_for_files(btme=1, caution_message=False):
                    self.state = 'completed'
                    self.disappeared = True
                    _logme.log('Job never appeared in the queue, but '
                               'outfiles still exist, assuming completion.',
                               'warn')
        if self.done and fetch_info:
            if self._wait_for_files(btme=1, caution_message=False):
                if not self._got_exitcode:
                    self.get_exitcode(update=False)
                if not self._got_times:
                    self.get_times(update=False)
        self._updating = False

    def _wait_for_files(self, btme=None, caution_message=False):
        """Block until files appear up to 'file_block_time' in config file.

        Aborts after 2 seconds if job exit code is not 0.

        Parameters
        ----------
        btme : int, optional
            Number of seconds to try for before giving up, default set in
            config file.
        caution_message : bool, optional
            Display a message if this is taking a while.

        Returns
        -------
        bool
            True if files found
        """
        if self._found_files:
            _logme.log('Already found files, not waiting again', 'debug')
            return True
        wait_time = 0.1 # seconds
        if btme:
            lvl = 'debug'
        else:
            lvl = 'warn'
            btme = _conf.get_option('jobs', 'file_block_time', 30)
        start = _dt.now()
        dsp   = False
        _logme.log('Checking for output files', 'debug')
        while True:
            runtime = (_dt.now() - start).seconds
            if caution_message and runtime > 1:
                _logme.log('Job complete.', 'info')
                _logme.log('Waiting for output files to appear.', 'info')
                caution_message = False
            if not dsp and runtime > 20:
                _logme.log('Still waiting for output files to appear',
                           'info')
                dsp = True
            count = 0
            outfiles = self.incomplete_outfiles
            tlen = len(outfiles)
            if not outfiles:
                _logme.log('No incomplete outfiles, assuming all found in ' +
                           '{} seconds'.format(runtime), 'debug')
                break
            for i in outfiles:
                if _os.path.isfile(i):
                    count += 1
            if count == tlen:
                _logme.log('All output files found in {} seconds'
                           .format(runtime), 'debug')
                break
            _sleep(wait_time)
            if runtime > btme:
                _logme.log('Job files have not appeared for ' +
                           '>{} seconds'.format(btme), lvl)
                return False
            if not self._updating:
                self.update()
            if runtime > 2 and self.get_exitcode(update=False) != 0:
                _logme.log('Job failed with exit code {}.'
                           .format(self.exitcode) + ' Cannot find files.',
                           'error')
                return False
        self._found_files = True
        return True

    def _update_name(self, name=None):
        """Make sure the job name is unique.

        Sets
        ----
        self.name

        Parameters
        ----------
        name : str, optional
            A name override, if no provided self.name used

        Returns
        -------
        name : str
        """
        # Set name
        name = name if name else self.name
        if not name:
            if callable(self.command):
                strcmd = str(self.command).strip('<>')
                parts = strcmd.split(' ')
                if parts[0] == 'bound':
                    name = '_'.join(parts[2:3])
                else:
                    parts.remove('function')
                    try:
                        parts.remove('built-in')
                    except ValueError:
                        pass
                    name = parts[0]
            else:
                name = self.command.split(' ')[0].split('/')[-1]

        # Make sure name not in queue
        if '.' not in name or not name.split('.')[-1] == self.uuid:
            name = '{}.{}'.format(name, self.uuid)

        self.name = name

        return name

    def __repr__(self):
        """Return simple job information."""
        outstr = "Job:{name}<{mode}:{qtype}".format(
            name=self.name, mode=self.kind, qtype=self.qtype)
        if self.submitted:
            outstr += ':{}'.format(self.id)
        outstr += "(command:{cmnd})".format(cmnd=self.command)
        if self.submitted or self.done:
            outstr += self.state.upper()
        elif self.written:
            outstr += "WRITTEN"
        else:
            outstr += "NOT_SUBMITTED"
        outstr += ">"
        return outstr

    def __str__(self):
        """Print job name and ID + status."""
        if not self._updating:
            self.update()
        return "Job: {name} ID: {id}, state: {state}".format(
            name=self.name, id=self.id, state=self.state)

    def __int__(self):
        """Return integer of ID."""
        if self.id:
            if str(self.id.isdigit()):
                return int(id)
        _logme.log('No ID yet.', 'error')
        return 0



# -*- coding: utf-8 -*-
"""
Wrapper functions to make some common jobs easier.

Functions
---------
submit
    Submit a script to the cluster
make_job
    Make a job compatible with the chosen cluster but do not submit
make_job_file
    Make a job file compatible with the chosen cluster.
clean
    Delete all files in jobs list or single Job object.
submit_file
    Submit an existing job file to the cluster.
clean_work_dirs
    Clean all files in the scriptpath and outpath directories.
clean_dir
    Delete all files made by this module in directory.
wait
    Wait for jobs to finish.
get
    Get results of jobs when they complete.
"""
import os  as _os
import sys as _sys
from time import sleep as _sleep
from subprocess import CalledProcessError as _CalledProcessError

###############################################################################
#                               Import Ourself                                #
###############################################################################


from . import run   as _run
from . import conf  as _conf
from . import queue as _queue
from . import logme as _logme
from . import batch_systems as _batch
from . import ClusterError as _ClusterError
from .job import Job

__all__ = ['submit', 'make_job', 'make_job_file', 'submit_file', 'clean_dir',
           'clean_work_dirs', 'clean', 'wait', 'get']

###############################################################################
#                            Submission Functions                             #
###############################################################################


def submit(command, args=None, kwargs=None, name=None, qtype=None,
           profile=None, **kwds):
    """Submit a script to the cluster.

    Parameters
    ----------
    command : function/str
        The command or function to execute.
    args : tuple/dict
        Optional arguments to add to command, particularly useful for
        functions.
    kwargs : dict
        Optional keyword arguments to pass to the command, only used for
        functions.
    name : str
        Optional name of the job. If not defined, guessed. If a job of the same
        name is already queued, an integer job number (not the queue number)
        will be added, ie.  <name>.1
    qtype : str
        Override the default queue type
    profile : str
        The name of a profile saved in the conf
    kwds
        *All other keywords are parsed into cluster keywords by the options
        system.* For available keywords see `fyrd.option_help()`

    Returns
    -------
    Job object
    """

    _batch.check_queue()  # Make sure the queue.MODE is usable

    job = Job(command=command, args=args, kwargs=kwargs, name=name,
              qtype=qtype, profile=profile, **kwds)

    job.write()
    job.submit()
    job.update()

    return job


#########################
#  Job file generation  #
#########################


def make_job(command, args=None, kwargs=None, name=None, qtype=None,
             profile=None, **kwds):
    """Make a job compatible with the chosen cluster but do not submit.

    Parameters
    ----------
    command : function/str
        The command or function to execute.
    args : tuple/dict
        Optional arguments to add to command, particularly useful for
        functions.
    kwargs : dict
        Optional keyword arguments to pass to the command, only used for
        functions.
    name : str
        Optional name of the job. If not defined, guessed. If a job of the same
        name is already queued, an integer job number (not the queue number)
        will be added, ie.  <name>.1
    qtype : str
        Override the default queue type
    profile : str
        The name of a profile saved in the conf
    kwds
        *All other keywords are parsed into cluster keywords by the options
        system.* For available keywords see `fyrd.option_help()`

    Returns
    -------
    Job object
    """

    _batch.check_queue()  # Make sure the queue.MODE is usable

    job = Job(command=command, args=args, kwargs=kwargs, name=name,
              qtype=qtype, profile=profile, **kwds)

    # Return the path to the script
    return job


def make_job_file(command, args=None, kwargs=None, name=None, qtype=None,
                  profile=None, **kwds):
    """Make a job file compatible with the chosen cluster.

    Parameters
    ----------
    command : function/str
        The command or function to execute.
    args : tuple/dict
        Optional arguments to add to command, particularly useful for
        functions.
    kwargs : dict
        Optional keyword arguments to pass to the command, only used for
        functions.
    name : str
        Optional name of the job. If not defined, guessed. If a job of the same
        name is already queued, an integer job number (not the queue number)
        will be added, ie.  <name>.1
    qtype : str
        Override the default queue type
    profile : str
        The name of a profile saved in the conf
    kwds
        *All other keywords are parsed into cluster keywords by the options
        system.* For available keywords see `fyrd.option_help()`

    Returns
    -------
    str
        Path to job file
    """

    _batch.check_queue()  # Make sure the queue.MODE is usable

    job = Job(command=command, args=args, kwargs=kwargs, name=name,
              qtype=qtype, profile=profile, **kwds)

    job = job.write()

    # Return the path to the script
    return job.submission.file_name


##############
#  Cleaning  #
##############


def clean(jobs, clean_outputs=False):
    """Delete all files in jobs list or single Job object.

    Parameters
    ----------
    jobs : fyrd.job.Job or list of fyrd.job.Job
        Job objects to clean
    clean_outputs : bool
        Also clean outputs.
    """
    jobs = _run.listify(jobs)
    for job in jobs:
        job.clean(delete_outputs=clean_outputs)


###############################################################################
#                      Job Object Independent Functions                       #
###############################################################################


def submit_file(script_file, dependencies=None, qtype=None, submit_args=None):
    """Submit an existing job file to the cluster.

    This function is independent of the Job object and just submits a file
    using a cluster appropriate method.

    Parameters
    ----------
    script_file : str
        The path to the file to submit
    dependencies: str or list of strings, optional
        A job number or list of job numbers to depend on
    qtype : str, optional
        The name of the queue system to use, auto-detected if not given.
    submit_args : dict
        A dictionary of keyword arguments for the submission script.

    Returns
    -------
    job_number : str
    """
    qtype = qtype if qtype else _batch.get_cluster_environment()
    _batch.check_queue(qtype)
    dependencies = _run.listify(dependencies)

    batch = _batch.get_batch_system(qtype)
    return batch.submit(script_file, dependencies)


def clean_work_dirs(outputs=False, confirm=False):
    """Clean all files in the scriptpath and outpath directories.

    Parameters
    ----------
    outputs : bool
        Also delete output files.
    confirm : bool
        Confirm on command line before deleteing.

    Returns
    -------
    files : list
        A list of deleted files.
    """
    files = []
    scriptpath = _conf.get_option('jobs', 'scriptpath')
    outpath    = _conf.get_option('jobs', 'outpath')
    if scriptpath:
        files += clean_dir(scriptpath, delete_outputs=outputs, confirm=confirm)
    if outputs and outpath and outpath != scriptpath:
        files += clean_dir(outpath, delete_outputs=True, confirm=confirm)
    return files


def clean_dir(directory=None, suffix=None, qtype=None, confirm=False,
              delete_outputs=None):
    """Delete all files made by this module in directory.

    CAUTION: The clean() function will delete **EVERY** file with
             extensions matching those these::
                 .<suffix>.err
                 .<suffix>.out
                 .<suffix>.out.func.pickle
                 .<suffix>.sbatch & .<suffix>.script for slurm mode
                 .<suffix>.qsub for torque mode
                 .<suffix>.job for local mode
                 _func.<suffix>.py
                 _func.<suffix>.py.pickle.in
                 _func.<suffix>.py.pickle.out

    .. note:: This function will change in the future to use batch system
              defined paths.

    Parameters
    ----------
    directory : str
        The directory to run in, defaults to the current directory.
    suffix : str
        Override the default suffix.
    qtype : str
        Only run on files of this qtype
    confirm : bool
        Ask the user before deleting the files
    delete_outputs : bool
        Delete all output files too.

    Returns
    -------
    list
        A set of deleted files
    """
    _batch.check_queue(qtype)  # Make sure the queue.MODE is usable

    if delete_outputs is None:
        delete_outputs = _conf.get_option('jobs', 'clean_outputs')

    # Sanitize arguments
    directory = _os.path.abspath(directory if directory else '.')
    suffix = suffix if suffix else _conf.get_option('jobs', 'suffix')

    # Extension patterns to delete
    extensions = ['_func.' + suffix + '.py']
    if delete_outputs:
        extensions += ['.' + suffix + '.err', '.' + suffix + '.out',
                       '_func.' + suffix + '.py.pickle.out',
                       '.' + suffix + '.out.func.pickle',
                       '.' + suffix + '.job']

    if qtype:
        if qtype == 'local':
            extensions.append('.' + suffix)
        elif qtype == 'slurm':
            extensions += ['.' + suffix + '.sbatch', '.' + suffix + '.script']
        elif qtype== 'torque':
            extensions.append('.' + suffix + '.qsub')
    else:
        extensions.append('.' + suffix)
        extensions.append('_func.' + suffix + '.py.pickle.in')
        extensions += ['.' + suffix + '.sbatch', '.' + suffix + '.script']
        extensions.append('.' + suffix + '.qsub')

    files = [_os.path.join(directory, i) for i in _os.listdir(directory)]
    files = [i for i in files if _os.path.isfile(i)]

    if not files:
        _logme.log('No files found.', 'debug')
        return []

    deleted = []
    for f in files:
        for extension in extensions:
            if f.endswith(extension):
                deleted.append(f)

    deleted = sorted(deleted)
    delete  = False

    if confirm:
        if deleted:
            prompt = [_os.path.basename(i) for i in deleted]
            _sys.stdout.write('Directory: {}\n'.format(directory))
            _sys.stdout.write('Files to delete::\n\t')
            _sys.stdout.write('\n\t'.join(prompt) + '\n')
            answer = _run.get_input("Do you want to delete these files? [Y/n]",
                                    'yesno', 'y')
            if answer == 'y':
                delete  = True
                _sys.stdout.write('Deleting...\n')
            else:
                _sys.stdout.write('Aborting\n')
                delete  = False
                deleted = []
        else:
            _sys.stdout.write('No files to delete.\n')
    else:
        delete = True

    if delete and deleted:
        for f in deleted:
            _os.remove(f)
        if confirm:
            _sys.stdout.write('Done\n')

    return deleted


###############################################################################
#               Simple Wrapper to Wait on Queue and Get Outputs               #
###############################################################################


def wait(jobs, notify=True):
    """Wait for jobs to finish.

    Parameters
    ----------
    jobs : fyrd.job.Job or str or list of either (mixed list fine)
        A single job or list of jobs, either Job objects or job numbers
    notify : str, True, or False, optional
        If True, both notification address and wait_time must be set in
        the [notify] section of the config. A notification email will be
        sent if the time exceeds this time. This is the default.
        If a string is passed, notification is forced and the string must
        be the to address.
        False means no notification

    Returns
    -------
    success : bool
        True if all jobs successful, false otherwise
    """
    q = _queue.Queue()
    return q.wait(jobs, notify=notify)


def get(jobs):
    """Get results of jobs when they complete.

    Parameters
    ----------
    jobs : fyrd.job.Job or list of fyrd.job.Job

    Returns
    -------
    list
        Outputs (STDOUT or return value) of jobs

    .. note:: This function also modifies the input Job objects, so they will
              contain all outputs and state information.
    """
    q = _queue.Queue()
    return q.get(jobs)
